# -*- coding: utf-8 -*-

from base64 import b64encode
from ipaddress import ip_network
from kvirt import common
from kvirt.defaults import IMAGES, METADATA_FIELDS
from kvirt.common import error, warning, pprint, success
from azure.identity import ClientSecretCredential
from azure.mgmt.authorization import AuthorizationManagementClient
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.compute.models import DiskCreateOption
from azure.mgmt.marketplaceordering import MarketplaceOrderingAgreements
from azure.mgmt.network.models import SecurityRule
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.storage import StorageManagementClient
from azure.mgmt.msi import ManagedServiceIdentityClient
from azure.mgmt.dns import DnsManagementClient
from azure.storage.blob import BlobServiceClient, BlobSasPermissions, generate_container_sas
from datetime import datetime, timezone, timedelta
import os
from random import choice
from string import ascii_letters, digits
from time import sleep
import webbrowser


def valid_password(password):
    return len(password) >= 8 and any(not c.isalnum() for c in password)\
        and any(c.isupper() for c in password) and any(c.islower() for c in password)


class Kazure(object):
    def __init__(self, subscription_id, tenant_id, app_id, secret, location='westus', resource_group='kcli',
                 admin_user='superadmin', admin_password=None, mail=None, storage_account=None, debug=False):
        credentials = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
        self.tenant_id = tenant_id
        self.app_id = app_id
        self.secret = secret
        self.subscription_id = subscription_id
        self.resource_client = ResourceManagementClient(credentials, subscription_id)
        self.compute_client = ComputeManagementClient(credentials, subscription_id)
        self.conn = self.compute_client
        self.network_client = NetworkManagementClient(credentials, subscription_id)
        self.agreements = MarketplaceOrderingAgreements(credentials, subscription_id)
        try:
            self.resource_client.resource_groups.create(resource_group, {'location': location})
        except:
            pass
        self.location = location
        self.resource_group = resource_group
        self.debug = debug
        self.admin_user = admin_user
        if admin_password is not None and not valid_password(admin_password):
            error("admin_password doesn't comply with policy. It needs at least 8chars and a special character")
            self.conn = None
            return
        else:
            self.admin_password = admin_password
        self.mail = mail
        if storage_account is not None:
            self.blob_service_client = BlobServiceClient(f"https://{storage_account}.blob.core.windows.net",
                                                         credential=credentials)
            storage_client = StorageManagementClient(credentials, subscription_id)
            self.storage_location = storage_client.storage_accounts.get_properties(resource_group,
                                                                                   storage_account).primary_location
            self.account_key = storage_client.storage_accounts.list_keys(resource_group, storage_account).keys[0].value
            self.storage_account = storage_account
        self.dns_client = DnsManagementClient(credentials, subscription_id)
        self.msi_client = ManagedServiceIdentityClient(credentials, subscription_id)
        self.auth_client = AuthorizationManagementClient(credentials, subscription_id)

    def close(self):
        print("not implemented")

    def exists(self, name):
        try:
            self.compute_client.virtual_machines.get(self.resource_group, name)
            return True
        except:
            return False

    def net_exists(self, name):
        networks, subnets = [], []
        for network in self.network_client.virtual_networks.list(self.resource_group):
            networks.append(network.name)
            for subnet in self.network_client.subnets.list(self.resource_group, network.name):
                subnets.append(os.path.basename(subnet.id))
        return name in networks + subnets

    def disk_exists(self, pool, name):
        return name in [os.path.basename(d.name) for d in self.compute_client.disks.list(self.resource_group)]

    def create(self, name, virttype=None, profile='', flavor=None, plan='kvirt', cpumodel='host-model',
               cpuflags=[], cpupinning=[], numcpus=2, memory=512, guestid='guestrhel764', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True, diskinterface='virtio', nets=['default'], iso=None,
               vnc=True, cloudinit=True, reserveip=False, reservedns=False, reservehost=False, start=True, keys=[],
               cmds=[], ips=None, netmasks=None, gateway=None, nested=True, dns=None, domain=None, tunnel=False,
               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[], cmdline=None,
               placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False, numamode=None, numa=[],
               pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[], vmuser=None, guestagent=True):
        if self.exists(name):
            return {'result': 'failure', 'reason': f"VM {name} already exists"}
        if flavor is None:
            matching = [f for f in self.list_flavors() if f[1] >= numcpus and f[2] >= memory]
            if matching:
                flavor = matching[0][0]
                pprint(f"Using flavor {flavor}")
            else:
                return {'result': 'failure', 'reason': 'Couldnt find flavor matching requirements'}
        elif flavor not in [f[0] for f in self.list_flavors()]:
            return {'result': 'failure', 'reason': f'Flavor {flavor} not found'}
        compute_client = self.compute_client
        network_client = self.network_client
        if image is None:
            return {'result': 'failure', 'reason': 'An image (or urn) is required'}
        elif image in IMAGES:
            publisher, offer, sku, version = self.__evaluate_image(image)
            if publisher is None:
                return {'result': 'failure', 'reason': 'Only centos, rhel, suse and ubuntu images are supported'}
        elif ':' in image and image.count(':') == 2:
            publisher, offer, sku = image.split(':')
            version = 'latest'
        elif image in self.volumes():
            publisher, offer, sku, version = None, None, None, None
        else:
            return {'result': 'failure', 'reason': f'Invalid image {image}'}
        need_agreement = False
        image_product, image_plan = None, None
        if publisher is not None:
            try:
                images = compute_client.virtual_machine_images.list(self.location, publisher, offer, sku)
                plan_id = os.path.basename(images[0].id)
                image_info = compute_client.virtual_machine_images.get(self.location, publisher, offer, sku, plan_id)
                if image_info.plan is not None:
                    need_agreement = True
                    image_product, image_plan = image_info.plan.product, image_info.plan.name
            except Exception as e:
                return {'result': 'failure', 'reason': f'Hit {e}'}
        if need_agreement:
            agreement = self.agreements.marketplace_agreements.get(publisher_id=publisher, offer_id=offer,
                                                                   plan_id=image_plan, offer_type='virtualmachine')
            if not agreement.accepted:
                parameters = {'publisher': publisher, 'product': image_product, 'plan': image_plan,
                              'license_text_link': agreement.license_text_link,
                              'privacy_policy_link': agreement.privacy_policy_link,
                              'marketplace_terms_link': agreement.marketplace_terms_link,
                              'retrieve_datetime': agreement.retrieve_datetime,
                              'signature': agreement.signature, 'accepted': True}
                self.agreements.marketplace_agreements.create(publisher_id=publisher, offer_id=offer,
                                                              plan_id=image_plan, offer_type='virtualmachine',
                                                              parameters=parameters)
        tags = {}
        for entry in [field for field in metadata if field in METADATA_FIELDS]:
            tags[entry] = metadata[entry]
        if self.admin_password is not None:
            admin_password = self.admin_password
        else:
            characters = ascii_letters + digits + '#%$+-?|!'
            admin_password = [choice(ascii_letters)] + [choice(digits)] + [choice(characters) for i in range(6)]
            admin_password = ''.join(admin_password).capitalize()
            pprint(f"Using admin_password {admin_password}")
        disk_size = int(disks[0]['size'] if isinstance(disks[0], dict) else disks[0])
        os_disk = {'name': f'{name}-disk-0', 'create_option': 'FromImage', 'delete_option': 'Delete'}
        if disk_size > 30:
            os_disk['disk_size_gb'] = disk_size
        storage_profile = {'os_disk': os_disk}
        if ':' in image:
            storage_profile.update({'image_reference': {'publisher': publisher, 'offer': offer, 'sku': sku,
                                                        'version': version}})

        else:
            image_id = self.compute_client.images.get(self.resource_group, image).id
            storage_profile.update({'image_reference': {'id': image_id}})
        data = {'location': self.location, 'os_profile': {'computer_name': name, 'admin_username': self.admin_user,
                                                          'admin_password': admin_password},
                'tags': tags,
                'hardware_profile': {'vm_size': flavor},
                'diagnostics_profile': {'boot_diagnostics': {'enabled': True, 'storage_uri': None}},
                'storage_profile': storage_profile}
        zone = overrides.get('az') or overrides.get('availability_zone') or overrides.get('zone')
        if zone is not None:
            data['zones'] = [str(zone)]
        if need_agreement:
            data['plan'] = {'publisher': publisher, 'name': offer, 'product': sku}
        sg_data = {'id': f"{name}-nsg", 'location': self.location}
        sg = network_client.network_security_groups.begin_create_or_update(self.resource_group, f"{name}-nsg", sg_data)
        sg = sg.result()
        openshift_node = 'kubetype' in metadata and metadata['kubetype'] == "openshift"
        if openshift_node:
            for index, port in enumerate([80, 443, 2379, 2380, 4789, 8080, 5443, 6081, 6443, 8443, 22624]):
                cluster = metadata['kube']
                if cluster not in self.list_security_groups():
                    self.create_security_group(f"{cluster}-nsg")
                rule_data = SecurityRule(protocol='Tcp', source_address_prefix='*',
                                         destination_address_prefix='*', access='Allow',
                                         direction='Inbound', description=f'tcp {port}',
                                         source_port_range='*', destination_port_ranges=[f"{port}"],
                                         priority=101 + index, name=f"tcp-{port}")
                network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                     f"tcp-{port}", rule_data)
            rule_data = SecurityRule(protocol='Udp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='udp 4789',
                                     source_port_range='*', destination_port_ranges=["4789"],
                                     priority=112, name="udp-4789")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "udp-4789", rule_data)
            rule_data = SecurityRule(protocol='Udp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='udp 6081',
                                     source_port_range='*', destination_port_ranges=["6081"],
                                     priority=113, name="udp-6081")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "udp-6081", rule_data)
            rule_data = SecurityRule(protocol='Tcp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='tcp 30000-32767',
                                     source_port_range='*', destination_port_ranges=["30000", "32767"],
                                     priority=114, name="tcp-30000-32767")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "tcp-30000-32767", rule_data)
            rule_data = SecurityRule(protocol='Udp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='udp 30000-32767',
                                     source_port_range='*', destination_port_ranges=["30000", "32767"],
                                     priority=115, name="udp-30000-32767")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "udp-30000-32767", rule_data)
            rule_data = SecurityRule(protocol='Tcp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='udp 10250-10259',
                                     source_port_range='*', destination_port_ranges=["10250", "10259"],
                                     priority=116, name="tcp-10250-10259")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "tcp-10250-10259", rule_data)
            rule_data = SecurityRule(protocol='Tcp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='tcp 9000-9999',
                                     source_port_range='*', destination_port_ranges=["9000", "9999"],
                                     priority=117, name="tcp-9000-9999")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "tcp-9000-9999", rule_data)
            rule_data = SecurityRule(protocol='Udp', source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description='udp 9000-9999',
                                     source_port_range='*', destination_port_ranges=["9000", "9999"],
                                     priority=118, name="udp-9000-9999")
            network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                 "udp-9000-9999", rule_data)
            msi_client = self.msi_client
            auth_client = self.auth_client
            identities = [i.name for i in msi_client.user_assigned_identities.list_by_subscription()]
            if cluster not in identities:
                identity_data = {'location': self.location}
                identity = msi_client.user_assigned_identities.create_or_update(self.resource_group, f"kcli-{cluster}",
                                                                                identity_data)
                principal_id = identity.principal_id
                scope = f"/subscriptions/{self.subscription_id}/resourceGroups/{self.resource_group}"
                role_id = [d.id for d in auth_client.role_definitions.list(scope) if d.role_name == 'Contributor'][0]
                role_data = {'role_definition_id': role_id, 'principal_id': principal_id,
                             'principal_type': 'ServicePrincipal'}
                auth_client.role_assignments.create(scope, principal_id, role_data)
            identity = f'/subscriptions/{self.subscription_id}/resourceGroups/{self.resource_group}/providers/'
            identity += f'Microsoft.ManagedIdentity/userAssignedIdentities/kcli-{cluster}'
            data['identity'] = {'type': 'userAssigned', 'userAssignedIdentities': {identity: {}}}
        network_interfaces = []
        subnets = self.list_subnets()
        for index, net in enumerate(nets):
            nic_name = f'{name}-eth{index}'
            netpublic = overrides.get('public', True)
            public_ip = None
            ip = None
            if isinstance(net, str):
                netname = net
                alias = []
            elif isinstance(net, dict) and 'name' in net:
                netname = net['name']
                ip = net.get('ip')
                alias = net.get('alias', [])
                if 'public' in net:
                    netpublic = net.get('public')
            matching_subnets = [sub for sub in subnets if sub == netname or subnets[sub]['network'] == netname]
            if matching_subnets:
                subnet = subnets[matching_subnets[0]]
                subnet_id = subnet['id']
                subnet_cidr = subnet['cidr']
                ip_configuration = {'name': nic_name, 'subnet': {'id': subnet_id}}
                if ip is not None:
                    ip_configuration['private_ip_allocation_method'] = 'Static'
                    ip_configuration['private_ip_address'] = ip
                ip_configurations = [ip_configuration]
                if ':' in subnet_cidr or subnet.get('dual_cidr') is not None:
                    ip_configuration = {'name': f'{nic_name}-ipv6', 'subnet': {'id': subnet_id},
                                        'private_ip_address_version': 'IPv6'}
                    ip_configurations.append(ip_configuration)
                nic_data = {'location': self.location, 'ip_configurations': ip_configurations}
                if index == 0:
                    nic_data['network_security_group'] = {"id": sg.id}
                    if netpublic:
                        ip_data = {"location": self.location, "sku": {"name": "Standard"},
                                   "public_ip_allocation_method": "Static", "public_ip_address_version": "IPV4"}
                        ip = network_client.public_ip_addresses.begin_create_or_update(self.resource_group,
                                                                                       f'{name}-ip', ip_data)

                        public_ip = ip.result()
                        rule_data = SecurityRule(protocol='Tcp', source_address_prefix='*',
                                                 destination_address_prefix='*', access='Allow',
                                                 direction='Inbound', description='tcp 22', source_port_range='*',
                                                 destination_port_ranges=["22"], priority=100, name="tcp-22")
                        network_client.security_rules.begin_create_or_update(self.resource_group, f"{name}-nsg",
                                                                             "tcp-22", rule_data)
                        nic_data['ip_configurations'][0]['public_ip_address'] = {"id": public_ip.id,
                                                                                 'delete_option': 'Delete'}
                nic = network_client.network_interfaces.begin_create_or_update(self.resource_group, nic_name, nic_data)
                nic_id = nic.result().id
                nic_reference = {'id': nic_id, 'delete_option': 'Delete', 'primary': True if index == 0 else False}
                network_interfaces.append(nic_reference)
            else:
                return {'result': 'failure', 'reason': f'Subnet {netname} not found'}
        data['network_profile'] = {'network_interfaces': network_interfaces}
        data_disks = []
        for index, disk in enumerate(disks):
            disk_name = f'{name}-disk{index}'
            if index == 0:
                continue
            lun = 10 + index
            if isinstance(disk, int):
                disk_size = disk
            elif isinstance(disk, str) and disk.isdigit():
                disk_size = int(disk)
            elif isinstance(disk, dict):
                disk_size = int(disk.get('size', 10))
            disk_data = {'location': self.location, 'disk_size_gb': disk_size,
                         'creation_data': {'create_option': DiskCreateOption.empty}}
            data_disk = compute_client.disks.begin_create_or_update(self.resource_group, disk_name, disk_data)
            data_disk_id = data_disk.result().id
            new_disk = {'lun': lun, 'name': disk_name, 'create_option': DiskCreateOption.attach,
                        'managed_disk': {'id': data_disk_id}, 'delete_option': 'Delete'}
            data_disks.append(new_disk)
        if data_disks:
            data['storage_profile']['data_disks'] = data_disks
        if cloudinit:
            if image is not None and common.needs_ignition(image):
                version = common.ignition_version(image)
                userdata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                           domain=domain, files=files, enableroot=enableroot,
                                           overrides=overrides, version=version, plan=plan, image=image,
                                           vmuser=vmuser)
            else:
                userdata = common.cloudinit(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                            domain=domain, files=files, enableroot=enableroot,
                                            overrides=overrides, fqdn=True, storemetadata=storemetadata,
                                            vmuser=vmuser)[0]
            custom_data = b64encode(userdata.encode('utf-8')).decode('latin-1')
            data['os_profile']['custom_data'] = custom_data
        if self.debug:
            print(data)
        result = compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, data)
        if not openshift_node or 'bootstrap' in name:
            result.wait()
        if reservedns and domain is not None:
            self.reserve_dns(name, nets=nets, domain=domain, alias=alias, instanceid=name)
        if 'loadbalancer' in overrides:
            lb = network_client.load_balancers.list(self.resource_group, overrides['loadbalancer'])
            backend_id = lb.backend_address_pools[0].id
            rule = lb.inbound_nat_rules[0] if lb.inbound_nat_rules else lb.load_balancing_rules[0]
            ports = rule.frontend_port_range_start if lb.inbound_nat_rules else rule.frontend_port
            self.add_vm_to_loadbalancer(name, backend_id, ports)
            self.update_metadata(name, 'loadbalancer', lb, append=True)
        return {'result': 'success'}

    def start(self, name):
        try:
            result = self.compute_client.virtual_machines.begin_start(self.resource_group, name)
            result.wait()
            return {'result': 'success'}
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}

    def stop(self, name, soft=False):
        try:
            if soft:
                result = self.compute_client.virtual_machines.begin_power_off(self.resource_group, name)
            else:
                result = self.compute_client.virtual_machines.begin_deallocate(self.resource_group, name)
            result.wait()
            return {'result': 'success'}
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}

    def create_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def delete_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def list_snapshots(self, base):
        print("not implemented")
        return []

    def revert_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def restart(self, name):
        try:
            result = self.compute_client.virtual_machines.begin_restart(self.resource_group, name)
            result.wait()
            return {'result': 'success'}
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}

    def info_host(self):
        return {"location": self.location, 'vms': len(self.list())}

    def status(self, name):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        return os.path.basename(vm.instance_view.statuses[1].code)

    def list(self):
        vms = []
        instances = self.compute_client.virtual_machines.list(self.resource_group)
        for vm in instances:
            try:
                vms.append(self.info(os.path.basename(vm.id)))
            except:
                continue
        if not vms:
            return []
        return sorted(vms, key=lambda x: x['name'])

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        self.serialconsole(name, web=web)

    def serialconsole(self, name, web=False):
        if self.mail is None:
            error("Serial console requires to set mail in your configuration")
            return
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name)
        except:
            error(f"VM {name} not found")
            return
        if vm.diagnostics_profile is None or not vm.diagnostics_profile.boot_diagnostics.enabled:
            pprint(f"Enabling Boot diagnostics for VM {name} ")
            vm.diagnostics_profile = {'boot_diagnostics': {'enabled': True, 'storage_uri': None}}
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()
        if '@' in self.mail:
            user = f"{self.mail.split('@')[0]}{''.join(self.mail.split('@')[1].split('.')[:-1])}.onmicrosoft."
            user += self.mail.split('@')[1].split('.')[-1]
        else:
            user = f"{self.mail}.onmicrosoft.com"
        url = f"https://portal.azure.com/?quickstart=true#@{user}/"
        url += f"resource/subscriptions/{self.subscription_id}/resourceGroups/kcli/providers/"
        url += f"Microsoft.Compute/virtualMachines/{name}/serialConsole"
        if web:
            return url
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{url}" if os.path.exists("/i_am_a_container") else url
            pprint(msg)
        else:
            pprint(f"Opening url {url}")
            webbrowser.open(url, new=2, autoraise=True)

    def info(self, name, output='plain', fields=[], values=False, vm=None, debug=False):
        compute_client = self.compute_client
        network_client = self.network_client
        yamlinfo = {}
        try:
            vm = compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {}
        yamlinfo['name'] = vm.name
        status = vm.instance_view.statuses
        yamlinfo['status'] = os.path.basename(status[1].code) if len(status) > 1 else 'N/A'
        yamlinfo['id'] = vm.vm_id
        if vm.zones is not None:
            yamlinfo['az'] = vm.zones[0]
        hardware_profile = vm.hardware_profile
        flavor = hardware_profile.vm_size
        yamlinfo['flavor'] = flavor
        yamlinfo['numcpus'], yamlinfo['memory'] = [f for f in self.list_flavors() if f[0] == flavor][0][1:]
        ips = []
        nets = []
        for index, nic in enumerate(vm.network_profile.network_interfaces):
            device = os.path.basename(nic.id)
            nic_data = network_client.network_interfaces.get(self.resource_group, device)
            if self.debug:
                print(nic_data)
            mac = nic_data.mac_address
            for entry in nic_data.ip_configurations:
                ips.append(entry.private_ip_address)
            private_ip = nic_data.ip_configurations[0].private_ip_address
            public_ip = nic_data.ip_configurations[0].public_ip_address
            if public_ip is not None:
                public_ip = network_client.public_ip_addresses.get(self.resource_group,
                                                                   os.path.basename(public_ip.id)).ip_address
            if index == 0:
                yamlinfo['ip'] = public_ip or private_ip
            subnet_name = os.path.basename(nic_data.ip_configurations[0].subnet.id)
            nets.append({'device': device, 'mac': mac, 'net': subnet_name, 'type': private_ip})
        yamlinfo['nets'] = nets
        if len(ips) > 1:
            yamlinfo['ips'] = ips
        for key in vm.tags:
            if key in METADATA_FIELDS:
                yamlinfo[key] = vm.tags[key]
        storage_profile = vm.storage_profile
        image_reference = storage_profile.image_reference
        if image_reference.offer is not None:
            image = f"{image_reference.offer}-{image_reference.sku}-{image_reference.version}"
        else:
            image = os.path.basename(image_reference.id)
        yamlinfo['image'] = image
        yamlinfo['user'] = common.get_user(yamlinfo['image'])
        disks = []
        for index, disk in enumerate([storage_profile.os_disk] + storage_profile.data_disks):
            device, disksize, diskformat = f'disk{index}', disk.disk_size_gb, disk.caching
            drivertype, path = 'N/A', disk.name
            disks.append({'device': device, 'size': disksize, 'format': diskformat, 'type': drivertype, 'path': path})
        yamlinfo['disks'] = disks
        yamlinfo['creationdate'] = vm.time_created
        if debug:
            yamlinfo['debug'] = vm
        return yamlinfo

    def ip(self, name):
        network_client = self.network_client
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        device = os.path.basename(vm.network_profile.network_interfaces[0].id)
        nic_data = network_client.network_interfaces.get(self.resource_group, device)
        private_ip = nic_data.ip_configurations[0].private_ip_address
        public_address = nic_data.ip_configurations[0].public_ip_address
        if public_address is not None:
            public_address = network_client.public_ip_addresses.get(self.resource_group,
                                                                    os.path.basename(public_address.id)).ip_address
        return public_address or private_ip

    def volumes(self, iso=False):
        publishers = ['RedHat', 'Suse', 'Canonical', 'Debian', 'MicrosoftWindowsServer', 'OpenLogic']
        images = []
        if iso:
            return []
        images.extend([image.name for image in self.compute_client.images.list()])
        for publisher in publishers:
            for o in self.compute_client.virtual_machine_images.list_offers(self.location, publisher):
                offer = o.name
                for s in self.compute_client.virtual_machine_images.list_skus(self.location, publisher, offer):
                    images.append(f"{publisher}:{offer}:{s.name}")
        return sorted(images)

    def delete(self, name, snapshots=False):
        compute_client = self.compute_client
        network_client = self.network_client
        try:
            vm = compute_client.virtual_machines.get(self.resource_group, name)
        except:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        tags = vm.tags
        dnsclient, domain = tags.get('dnsclient'), tags.get('domain')
        result = compute_client.virtual_machines.begin_delete(self.resource_group, name, force_deletion=True)
        result.wait()
        try:
            network_client.network_security_groups.begin_delete(self.resource_group, f"{name}-nsg")
        except:
            pass
        if domain is not None and dnsclient is None:
            self.delete_dns(name, domain, name)
        return {'result': 'success'}

    def dnsinfo(self, name):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name)
        except:
            return None, None
        tags = vm.tags
        dnsclient, domain = tags.get('dnsclient'), tags.get('domain')
        return dnsclient, domain

    def clone(self, old, new, full=False, start=False):
        print("not implemented")

    def update_metadata(self, name, metatype, metavalue, append=False):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name)
        except:
            error(f"VM {name} not found")
            return 1
        if vm.tags is None:
            vm.tags = {metatype: metavalue}
        else:
            vm.tags[metatype] = metavalue
        self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)

    def update_memory(self, name, memory):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        current_numcpus, current_memory = [f for f in self.list_flavors() if f[0] == vm.hardware_profile.vm_size][0][1:]
        if memory > current_memory:
            matching = [f for f in self.list_flavors() if f[2] >= memory]
            if matching:
                flavor = matching[0][0]
                pprint(f"Using flavor {flavor}")
            else:
                return {'result': 'failure', 'reason': 'Couldnt find flavor matching requirements'}
            if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
                error(f"Can't update memory of VM {name} while up")
                return {'result': 'failure', 'reason': f"VM {name} up"}
            vm.hardware_profile.vm_size = flavor
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()

    def update_cpus(self, name, numcpus):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        current_numcpus, current_memory = [f for f in self.list_flavors() if f[0] == vm.hardware_profile.vm_size][0][1:]
        if numcpus > current_numcpus:
            matching = [f for f in self.list_flavors() if f[1] >= numcpus]
            if matching:
                flavor = matching[0][0]
                pprint(f"Using flavor {flavor}")
            else:
                return {'result': 'failure', 'reason': 'Couldnt find flavor matching requirements'}
            if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
                error(f"Can't update cpus of VM {name} while up")
                return {'result': 'failure', 'reason': f"VM {name} up"}
            vm.hardware_profile.vm_size = flavor
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()

    def update_start(self, name, start=True):
        print("not implemented")

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)

    def update_iso(self, name, iso):
        print("not implemented")

    def update_flavor(self, name, flavor):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if flavor not in [f[0] for f in self.list_flavors()]:
            return {'result': 'failure', 'reason': f'Flavor {flavor} not found'}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't update flavor of VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        vm.hardware_profile.vm_size = flavor
        result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
        result.wait()

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        data = {'location': self.location, 'disk_size_gb': size,
                'creation_data': {'create_option': DiskCreateOption.empty}}
        self.compute_client.disks.begin_create_or_update(self.resource_group, name, data)

    def add_disk(self, name, size=1, pool=None, thin=True, image=None, shareable=False, existing=None,
                 interface='virtio', novm=False, overrides={}):
        if novm:
            return self.create_disk(name=name, size=size, pool=pool, thin=thin, image=image)
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't add disk to VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        index = len(vm.storage_profile.data_disks) + 1
        disk_name = f"{name}-disk{index}"
        disk_data = {'location': self.location, 'disk_size_gb': size,
                     'creation_data': {'create_option': DiskCreateOption.empty}}
        data_disk = self.compute_client.disks.begin_create_or_update(self.resource_group, disk_name, disk_data)
        data_disk_id = data_disk.result().id
        lun = 10 + index
        new_disk = {'lun': lun, 'name': disk_name, 'create_option': DiskCreateOption.attach,
                    'managed_disk': {'id': data_disk_id}, 'delete_option': 'Delete'}
        vm.storage_profile.data_disks.append(new_disk)
        result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
        result.wait()

    def delete_disk(self, name, diskname, pool=None, novm=False):
        if novm:
            result = self.compute_client.disks.begin_delete(self.resource_group, diskname)
            result.wait()
            return {'result': 'success'}
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't delete disk from VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        disks = []
        for disk in vm.storage_profile.data_disks:
            if os.path.basename(disk.name).endswith(diskname):
                disk_name = os.path.basename(disk.name)
            else:
                disks.append(disk)
        if disks != vm.storage_profile.data_disks:
            vm.storage_profile.data_disks = disks
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()
            result = self.compute_client.disks.begin_delete(self.resource_group, disk_name)
            result.wait()

    def list_disks(self):
        disks = {}
        for disk in self.compute_client.disks.list():
            disks[disk.name] = {'pool': 'default', 'path': disk.name}
        return disks

    def detach_disks(self, name):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            result = self.compute_client.virtual_machines.begin_deallocate(self.resource_group, name)
            result.wait()
        if len(vm.storage_profile.data_disks) > 1:
            vm.storage_profile.data_disks = vm.storage_profile.data_disks[:1]
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()
        return {'result': 'success'}

    def add_nic(self, name, network, model='virtio'):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't add nic to VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        subnets = self.list_subnets()
        if network not in subnets:
            return {'result': 'failure', 'reason': f'Subnet {network} not found'}
        else:
            subnet_id = subnets[network]['az']
        index = len(vm.network_profile.network_interfaces)
        nic_name = f'{name}-eth{index}'
        nic_data = {'location': self.location, 'ip_configurations': [{'name': nic_name, 'subnet': {'id': subnet_id}}]}
        nic = self.network_client.network_interfaces.begin_create_or_update(self.resource_group, nic_name, nic_data)
        nic_id = nic.result().id
        nic_reference = {'id': nic_id, 'delete_option': 'Delete', 'primary': False}
        vm.network_profile.network_interfaces.append(nic_reference)
        result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
        result.wait()

    def delete_nic(self, name, interface):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't delete nic from VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        interfaces = []
        for nic in vm.network_profile.network_interfaces:
            if os.path.basename(nic.id).endswith(interface):
                nic_name = os.path.basename(nic.id)
            else:
                interfaces.append(nic)
        if interfaces != vm.network_profile.network_interfaces:
            vm.network_profile.network_interfaces = interfaces
            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
            result.wait()
            result = self.network_client.network_interfaces.begin_delete(self.resource_group, nic_name)
            result.wait()

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")

    def delete_image(self, image, pool=None):
        if image in [i.name for i in self.compute_client.images.list()]:
            result = self.compute_client.images.begin_delete(self.resource_group, image)
            result.wait()
            return {'result': 'success'}
        else:
            return {'result': 'failure', 'reason': f'Image {image} not found'}

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        bucket = None
        # if 'blob.core.windows.net' not in url:
        if self.storage_account not in url:
            bucket = f"kcli-import-{''.join(choice(digits) for _ in range(4))}"
            self.blob_service_client.create_container(bucket)
            url = self.upload_to_bucket(bucket, url)
        shortimage = os.path.basename(url).split('?')[0]
        image = name or shortimage
        image_data = {'location': self.storage_location,
                      'storage_profile': {'osDisk': {'osType': 'Linux', 'blobUri': url, 'osState': 'generalized'}},
                      'hyper_v_generation': 'V1'}
        result = self.compute_client.images.begin_create_or_update(self.resource_group, image, image_data)
        result.wait()
        succeeded = 'Nope'
        while succeeded != 'Succeeded':
            sleep(5)
            pprint(f"Waiting for image {image} to be available")
            succeeded = self.compute_client.images.get(self.resource_group, image).provisioning_state
        if bucket is not None:
            self.blob_service_client.delete_container(bucket)
        return {'result': 'success'}

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        if cidr is not None:
            try:
                network = ip_network(cidr, strict=False)
            except:
                return {'result': 'failure', 'reason': f"Invalid Cidr {cidr}"}
            network_ipv6 = str(network.version) == "6"
        else:
            return {'result': 'failure', 'reason': "Cidr needed"}
        dual_cidr = overrides.get('dual_cidr')
        if dual_cidr is not None:
            try:
                dual_network = ip_network(dual_cidr, strict=False)
            except:
                return {'result': 'failure', 'reason': f"Invalid Dual Cidr {cidr}"}
            dual_ipv6 = str(dual_network.version) == "6"
            if network_ipv6 == dual_ipv6:
                return {'result': 'failure', 'reason': "cidr and dual_cidr must be of different types"}
        elif network_ipv6:
            return {'result': 'failure', 'reason': "ipv6 requires dual_cidr to be set"}
        data = {'location': self.location, 'address_space': {'address_prefix': cidr, 'address_prefixes': [cidr]},
                'tags': {'plan': plan}}
        if dual_cidr is not None:
            data['address_space']['address_prefixes'].append(dual_cidr)
        result = self.network_client.virtual_networks.begin_create_or_update(self.resource_group, name, data)
        result.wait()
        ip_data = {'location': self.location, "sku": {"name": "Standard"},
                   "public_ip_allocation_method": "Static"}
        ip_data['public_ip_address_version'] = 'IPV6' if network_ipv6 and dual_cidr is None else 'IPV4'
        if not network_ipv6 or dual_cidr is not None:
            public_ip_name = f'network-{name}-ip'
            public_ip = self.network_client.public_ip_addresses.begin_create_or_update(self.resource_group,
                                                                                       public_ip_name, ip_data)
            public_ip = public_ip.result()
            public_ip_id = public_ip.id
            public_ip = public_ip.ip_address
            nat_gateway_data = {'location': self.location, "sku": {"name": "Standard"},
                                'public_ip_addresses': [{"id": public_ip_id, 'delete_option': 'Delete'}]}
            nat_gateway = self.network_client.nat_gateways.begin_create_or_update(self.resource_group, f'nat-{name}',
                                                                                  nat_gateway_data)
            nat_gateway = nat_gateway.result()
            nat_gateway_id = nat_gateway.id
        if overrides.get('create_subnet', True):
            pprint(f"Creating first subnet {name}-subnet1")
            subnet_cidr = overrides.get('subnet_cidr')
            if subnet_cidr is not None:
                try:
                    subnet = ip_network(subnet_cidr, strict=False)
                except:
                    return {'result': 'failure', 'reason': f"Invalid Cidr {subnet_cidr}"}
                if not subnet.subnet_of(network):
                    return {'result': 'failure', 'reason': f"{subnet_cidr} isnt part of {cidr}"}
            else:
                subnet_cidr = cidr
            subnet_data = {'address_prefix': subnet_cidr, 'address_prefixes': [subnet_cidr]}
            subnet_ipv6 = ':' in subnet_cidr
            if not nat and not subnet_ipv6:
                data['nat_gateway'] = {'id': nat_gateway_id}
            self.network_client.subnets.begin_create_or_update(self.resource_group, name, f'{name}-subnet1',
                                                               subnet_data)
        return {'result': 'success'}

    def delete_network(self, name=None, cidr=None, force=False):
        vms = self.network_ports(name)
        if vms:
            if not force:
                vms = ','.join(vms)
                return {'result': 'failure', 'reason': f"Network {name} is being used by the following vms: {vms}"}
            for vm in vms:
                self.delete(vm)
        result = self.network_client.virtual_networks.begin_delete(self.resource_group, name)
        result.wait()
        for n in self.network_client.nat_gateways.list(self.resource_group):
            if n.name == f'nat-{name}':
                pprint(f"Deleting nat_gateway nat-{name}")
                self.network_client.nat_gateways.begin_delete(self.resource_group, f'nat-{name}')
        return {'result': 'success'}

    def list_pools(self):
        return []

    def list_networks(self):
        resource_group = self.resource_group
        networks = {}
        network_client = self.network_client
        for network in network_client.virtual_networks.list(resource_group):
            networkname = network.name
            cidr = network_client.virtual_networks.get(resource_group, networkname).address_space.address_prefixes[0]
            dhcp = True
            domain = network.resource_guid
            mode = ''
            networks[networkname] = {'cidr': cidr, 'dhcp': dhcp, 'domain': domain, 'type': 'routed', 'mode': mode}
            plan = network.tags['plan'] if network.tags is not None and 'plan' in network.tags else 'N/A'
            networks[networkname]['plan'] = plan
        return networks

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        if self.debug and networkinfo:
            print(self.network_client.virtual_networks.get(self.resource_group, name))
        return networkinfo

    def info_subnet(self, name):
        subnets = self.list_subnets()
        if name in subnets:
            return subnets[name]
        else:
            msg = f"Subnet {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}

    def list_subnets(self):
        subnets = {}
        network_client = self.network_client
        for network in network_client.virtual_networks.list(self.resource_group):
            for subnet in network_client.subnets.list(self.resource_group, network.name):
                if self.debug:
                    print(subnet)
                address_prefixes = subnet.address_prefixes or [subnet.address_prefix]
                cidr = address_prefixes[0]
                subnet_id = subnet.id
                subnets[subnet.name] = {'cidr': cidr, 'id': subnet_id, 'network': network.name}
                if len(address_prefixes) > 1:
                    subnets[subnet.name]['dual_cidr'] = address_prefixes[1]
        return subnets

    def delete_pool(self, name, full=False):
        print("not implemented")

    def network_ports(self, name):
        results = []
        all_subnets = self.list_subnets()
        subnets = [s for s in all_subnets if all_subnets[s]['network'] == name]
        for nic in self.network_client.network_interfaces.list(self.resource_group):
            ip_configurations = nic.ip_configurations or []
            for ip_configuration in ip_configurations:
                subnet_id = ip_configuration.subnet.id
                if subnet_id is not None and [s for s in subnets if all_subnets[s]['id'] == subnet_id]:
                    new_name = os.path.basename(nic.virtual_machine.id) if nic.virtual_machine is not None else nic.name
                    results.append(new_name)
        return results

    def vm_ports(self, name):
        return ['default']

    def get_pool_path(self, pool):
        print("not implemented")

    def list_flavors(self):
        vm_sizes = self.compute_client.virtual_machine_sizes.list(self.location)
        return [[e.name, e.number_of_cores, e.memory_in_mb] for e in vm_sizes]

    def export(self, name, image=None):
        print("not implemented")

    def create_bucket(self, bucket, public=False):
        if bucket in self.list_buckets():
            error(f"Bucket {bucket} already there")
            return
        public_access = 'container' if public else None
        self.blob_service_client.create_container(bucket, public_access=public_access)

    def delete_bucket(self, bucket):
        if bucket not in self.list_buckets():
            error(f"Inexistent bucket {bucket}")
            return
        self.blob_service_client.delete_container(bucket)

    def delete_from_bucket(self, bucket, path):
        if bucket not in self.list_buckets():
            error(f"Bucket {bucket} doesn't exist")
            return
        container_client = self.blob_service_client.get_container_client(bucket)
        if not os.path.basename(path) in container_client.list_blob_names():
            error(f"Bucket file {path} not found in Bucket {bucket}")
            return
        else:
            container_client.delete_blob(os.path.basename(path))

    def download_from_bucket(self, bucket, path):
        if bucket not in self.list_buckets():
            error(f"Bucket {bucket} doesn't exist")
            return
        container_client = self.blob_service_client.get_container_client(bucket)
        with open(path, mode="wb") as download_file:
            blob_name = os.path.basename(path)
            download_file.write(container_client.download_blob(blob_name).readall())

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        file_path = os.path.expanduser(path)
        file_name = os.path.basename(path)
        if bucket not in self.list_buckets():
            error(f"Bucket {bucket} doesn't exist")
            return
        blob_client = self.blob_service_client.get_blob_client(bucket, file_name)
        if path.startswith('http'):
            blob_client.start_copy_from_url(path)
            status = 'pending'
            while status == 'pending':
                sleep(5)
                pprint("Waiting for copy to finish")
                copy_blob = blob_client.get_blob_properties()
                status = copy_blob['copy']['status']
            if status == 'success':
                success("Copy operation completed successfully.")
            else:
                error(f"Copy operation status: {status}")
        elif not os.path.exists(file_path):
            error(f"Path {path} doesn't exist")
            return
        else:
            with open(file=file_path, mode="rb") as data:
                blob_client.upload_blob(data, overwrite=True)
        return blob_client.url

    def list_buckets(self):
        return [bucket.name for bucket in self.blob_service_client.list_containers()]

    def list_bucketfiles(self, bucket):
        container_client = self.blob_service_client.get_container_client(bucket)
        return [blob.name for blob in container_client.list_blobs()]

    def public_bucketfile_url(self, bucket, path):
        start_time = datetime.now(timezone.utc)
        expiry_time = start_time + timedelta(days=1)
        account_key = self.account_key
        sas_token = generate_container_sas(account_name=self.storage_account, account_key=account_key,
                                           container_name=bucket, permission=BlobSasPermissions(read=True),
                                           expiry=expiry_time, start=start_time, protocol='https')
        sas_token = sas_token.replace('%3A', ':')
        return f"https://{self.storage_account}.blob.core.windows.net/{bucket}/{path}?{sas_token}"

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False, instanceid=None):
        dns_client = self.dns_client
        if domain is None:
            domain = nets[0]
        internalip = None
        pprint(f"Using domain {domain}")
        cluster = None
        fqdn = f"{name}.{domain}"
        if fqdn.split('-')[0] == fqdn.split('.')[1]:
            cluster = fqdn.split('-')[0]
            name = '.'.join(fqdn.split('.')[:1])
            domain = fqdn.replace(f"{name}.", '').replace(f"{cluster}.", '')
        if domain not in [os.path.basename(z.id) for z in dns_client.zones.list()]:
            error(f"Domain {domain} not found")
            return {'result': 'failure', 'reason': f"Domain {domain} not found"}
        entry = name if cluster is None else f"{name}.{cluster}"
        if cluster is not None and ('ctlplane' in name or 'worker' in name):
            counter = 0
            while counter != 100:
                internalip = self.internalip(name)
                if internalip is None:
                    sleep(5)
                    pprint(f"Waiting 5 seconds to grab internal ip and create DNS record for {name}")
                    counter += 10
                else:
                    break
        if ip is None:
            net = nets[0]
            if isinstance(net, dict):
                ip = net.get('ip')
            if ip is None:
                counter = 0
                while counter != 100:
                    ip = self.ip(instanceid)
                    if ip is None:
                        sleep(5)
                        pprint("Waiting 5 seconds to grab ip and create DNS record...")
                        counter += 10
                    else:
                        break
        if ip is None:
            error(f"Couldn't assign DNS for {name}")
            return
        dnsip = ip if internalip is None else internalip
        dns_data = {"ttl": 300, "arecords": [{"ipv4_address": dnsip}]}
        dns_client.record_sets.create_or_update(self.resource_group, domain, entry, 'A', dns_data)
        for a in alias:
            if a == '*':
                if cluster is not None and ('ctlplane' in name or 'worker' in name):
                    alias_entry = f'*.apps.{cluster}'
                else:
                    alias_entry = f'*.{name}'
            else:
                alias_entry = a.replace(f'.{domain}', '')
            dns_data = {"ttl": 300, "cname_Record": {'cname': f"{entry}.{domain}"}}
            dns_client.record_sets.create_or_update(self.resource_group, domain, alias_entry, 'CNAME', dns_data)
        return {'result': 'success'}

    def delete_dns(self, name, domain, allentries=False):
        cluster = None
        fqdn = f"{name}.{domain}"
        if fqdn.split('-')[0] == fqdn.split('.')[1]:
            cluster = fqdn.split('-')[0]
            name = '.'.join(fqdn.split('.')[:1])
            domain = fqdn.replace(f"{name}.", '').replace(f"{cluster}.", '')
        if domain not in [os.path.basename(z.id) for z in self.dns_client.zones.list()]:
            error(f"Domain {domain} not found")
            return {'result': 'failure', 'reason': f"Domain {domain} not found"}
        entry = name if cluster is None else f"{name}.{cluster}"
        found = False
        for record in self.dns_client.record_sets.list_all_by_dns_zone(self.resource_group, domain):
            if entry in record.name:
                found = True
                _type = os.path.basename(record.type)
                self.dns_client.record_sets.delete(self.resource_group, domain, record.name, _type)
            elif record.cname_record is not None and entry in str(record.cname_record):
                found = True
                _type = os.path.basename(record.type)
                self.dns_client.record_sets.delete(self.resource_group, domain, record.name, _type)
        if not found:
            error(f"Entry {entry} not found")
            return {'result': 'failure', 'reason': f"Entry {entry} not found"}
        else:
            return {'result': 'success'}

    def list_dns(self, domain):
        results = []
        if domain not in [os.path.basename(d.id) for d in self.dns_client.zones.list()]:
            error(f"Domain {domain} not found")
            return results
        for entry in self.dns_client.record_sets.list_all_by_dns_zone(self.resource_group, domain):
            name, _type, ttl = entry.name, os.path.basename(entry.type), entry.ttl
            data = ''
            if entry.a_records is not None:
                data = ' '.join([x.ipv4_address for x in entry.a_records])
            elif entry.aaaa_records is not None:
                data = ' '.join([x.ipv6_address for x in entry.aaaa_records])
            elif entry.soa_record is not None:
                data = entry.soa_record.host
            elif entry.ns_records is not None:
                data = ' '.join([x.nsdname for x in entry.ns_records])
            elif entry.cname_record is not None:
                data = entry.cname_record.cname
            elif entry.mx_records is not None:
                data = ' '.join([x.exchange for x in entry.mx_records])
            elif entry.ptr_records is not None:
                data = ' '.join([x.ptrdname for x in entry.ptr_records])
            elif entry.srv_records is not None:
                data = ' '.join([f"{x.priority} {x.weight} {x.port} {x.target}" for x in entry.srv_records])
            elif entry.txt_records is not None:
                data = ' '.join([x.value for x in entry.txt_records])
            elif entry.caa_records is not None:
                data = ' '.join([f"{x.flag} {x.tag} {x.value}" for x in entry.caa_records])
            results.append([name, _type, ttl, data])
        return results

    def update_nic(self, name, index, network):
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
        except:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        subnets = self.list_subnets()
        if network not in subnets:
            msg = f"Subnet {network} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        if len(vm.network_profile.network_interfaces) < index:
            msg = f"Nic {index} not found in VM {name}"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
            error(f"Can't update memory of VM {name} while up")
            return {'result': 'failure', 'reason': f"VM {name} up"}
        vm.network_profile.network_interfaces.ip_configurations[index].subnet.id = subnets[network]['az']
        result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
        result.wait()
        return {'result': 'success'}

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        return self.create_network(name, overrides=overrides)

    def list_security_groups(self, network=None):
        sgs = []
        network_client = self.network_client
        for sg in network_client.network_security_groups.list(self.resource_group):
            sgs.append(sg.name)
        return sgs

    def create_security_group(self, name, overrides={}):
        ports = overrides.get('ports', [])
        network_client = self.network_client
        sg_data = {'id': name, 'location': self.location}
        sg = network_client.network_security_groups.begin_create_or_update(self.resource_group, name, sg_data)
        sg = sg.result()
        for index, port in enumerate(ports):
            if isinstance(port, str) or isinstance(port, int):
                protocol = 'Tcp'
                destport = int(port)
            elif isinstance(port, dict):
                protocol = port.get('protocol', 'tcp').capitalize()
                destport = port.get('to')
                if destport is None:
                    warning(f"Missing to in {ports}. Skipping")
                    continue
            priority = 200 + index
            rule_data = SecurityRule(protocol=protocol, source_address_prefix='*',
                                     destination_address_prefix='*', access='Allow',
                                     direction='Inbound', description=f'port {destport}', source_port_range='*',
                                     destination_port_ranges=[destport], priority=priority, name=f"port-{destport}")
            network_client.security_rules.begin_create_or_update(self.resource_group, name, f"port-{destport}",
                                                                 rule_data)
        return {'result': 'success'}

    def delete_security_group(self, name):
        self.network_client.network_security_groups.begin_delete(self.resource_group, name)
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def __evaluate_image(self, image):
        if image.startswith('centos'):
            return 'OpenLogic', 'CentOS', '8_5', 'latest'
        elif image.startswith('rhel8'):
            return 'RedHat', 'RHEL', '8_8', 'latest'
        elif image.startswith('rhel9'):
            return 'RedHat', 'RHEL', '9_2', 'latest'
        elif image.startswith('rhcos'):
            return 'redhat-limited', 'rh-ocp-worker', 'rh-ocp-worker', 'latest'
        elif 'suse' in image:
            return 'SUSE', 'openSUSE-leap-15-4', 'gen2', 'latest'
        elif image.startswith('ubuntu'):
            return 'Canonical', '0001-com-ubuntu-server-jammy', '22_04-lts-gen2', 'latest'
        else:
            return None, None, None, None

    def create_loadbalancer(self, name, ports=[], checkpath='/index.html', vms=[], domain=None, checkport=80, alias=[],
                            internal=False, dnsclient=None, ip=None):
        dual = False
        if not vms:
            msg = "Creating a load balancer requires to specify vms"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        subnet_id, backend_pool = None, None
        for index, vm in enumerate(vms):
            info = self.info(vm)
            if not info:
                msg = f"Vm {vm} not found"
                return {'result': 'failure', 'reason': msg}
            loadbalancer = info.get('loadbalancer')
            if loadbalancer is not None:
                backend_pool = loadbalancer
            if index == 0 and internal:
                subnet = self.list_subnets()[info['nets'][0]['net']]
                subnet_id = subnet['id']
                dual = 'dual_cidr' in subnet
        ports = [int(port) for port in ports]
        ports = list(dict.fromkeys(ports))
        network_client = self.network_client
        ip_data = {"location": self.location, "sku": {"name": "Standard"},
                   "public_ip_allocation_method": "Static", "public_ip_address_version": "IPV4"}
        if not internal:
            public_ip = network_client.public_ip_addresses.begin_create_or_update(self.resource_group, f'{name}-ip',
                                                                                  ip_data)
            public_ip = public_ip.result()
            frontend_ip_configurations = [{'name': name, 'public_ip_address': {'id': public_ip.id}}]
        else:
            frontend_ip_configurations = [{'name': name, 'private_ip_address_allocation': 'Dynamic',
                                           'subnet': {'id': subnet_id}}]
            if dual:
                frontend_ip_configuration = {'name': f'{name}-ipv6', 'private_ip_address_allocation': 'Dynamic',
                                             'subnet': {'id': subnet_id}, 'private_ip_address_version': 'IPv6'}
                frontend_ip_configurations.append(frontend_ip_configuration)
        backend_name = backend_pool or name
        backend_address_pools = [{'name': backend_name}]
        if dual:
            backend_address_pools.append({'name': f'{backend_name}-ipv6'})
        checkport = ports[0]
        probes = [{'name': name, 'protocol': 'tcp', 'port': checkport, 'interval_in_seconds': 15,
                   'number_of_probes': 4}]
        lb_data = {'location': self.location,
                   'sku': {'name': 'Standard'},
                   'frontend_ip_configurations': frontend_ip_configurations,
                   'backend_address_pools': backend_address_pools,
                   'probes': probes}
        tags = {}
        if domain is not None:
            tags['domain'] = domain
        if dnsclient is not None:
            tags['dnsclient'] = dnsclient
        if tags:
            lb_data['tags'] = tags
        lb = network_client.load_balancers.begin_create_or_update(self.resource_group, name, lb_data).result()
        frontend_id = lb.frontend_ip_configurations[0].id
        backend_id = lb.backend_address_pools[0].id
        probe_id = lb.probes[0].id
        load_balancing_rules = [{'name': f"{name}-rule-{index}", 'protocol': 'tcp',
                                 'backend_address_pool': {'id': backend_id}, 'backend_port': port,
                                 'frontend_port': port, 'enable_floating_ip': False, 'idle_timeout_in_minutes': 4,
                                 'probe': {'id': probe_id},
                                 'frontend_ip_configuration': {'id': frontend_id}} for index, port in enumerate(ports)]
        backend_id_dual = None
        if dual:
            frontend_id = lb.frontend_ip_configurations[1].id
            backend_id_dual = lb.backend_address_pools[1].id
            dual_rules = [{'name': f"{name}-ipv6-rule-{index}", 'protocol': 'tcp',
                           'backend_address_pool': {'id': backend_id_dual}, 'backend_port': port,
                           'frontend_port': port, 'enable_floating_ip': False, 'idle_timeout_in_minutes': 4,
                           'probe': {'id': probe_id},
                           'frontend_ip_configuration': {'id': frontend_id}} for index, port in enumerate(ports)]
            load_balancing_rules.extend(dual_rules)
        lb.load_balancing_rules = load_balancing_rules
        lb = network_client.load_balancers.begin_create_or_update(self.resource_group, name, lb).result()
        if self.debug:
            print(lb)
        for index, vm in enumerate(vms):
            self.add_vm_to_loadbalancer(vm, backend_id, ports, backend_id_dual=backend_id_dual)
            self.update_metadata(vm, 'loadbalancer', name)
        if domain is not None:
            if not internal:
                public_ip_id = os.path.basename(public_ip.id)
                ip = network_client.public_ip_addresses.get(self.resource_group, public_ip_id).ip_address
            else:
                ip = lb.frontend_ip_configurations[0].private_ip_address
            self.reserve_dns(name, ip=ip, domain=domain, alias=alias)
        return {'result': 'success'}

    def delete_loadbalancer(self, name):
        network_client = self.network_client
        for lb in network_client.load_balancers.list(self.resource_group):
            if lb.name == name:
                domain, dnsclient = None, None
                tags = lb.tags or {}
                if 'dnsclient' in tags:
                    dnsclient = tags['dnsclient']
                if 'domain' in tags:
                    domain = tags['domain']
                    pprint(f"Using found domain {domain}")
                public_address = lb.frontend_ip_configurations[0].public_ip_address
                result = network_client.load_balancers.begin_delete(self.resource_group, name)
                result.wait()
                if public_address is not None:
                    result = network_client.public_ip_addresses.begin_delete(self.resource_group,
                                                                             os.path.basename(public_address.id))
                    result.wait()
                if domain is not None and dnsclient is None:
                    warning(f"Deleting DNS {name}.{domain}")
                    self.delete_dns(name, domain, name)
                elif dnsclient is not None:
                    return dnsclient
                if f"{name}-nsg" in self.list_security_groups():
                    network_client.network_security_groups.begin_delete(self.resource_group, f"{name}-nsg")
                return {'result': 'success'}
        error(f"Loadbalancer {name} not found")
        return {'result': 'success'}

    def list_loadbalancers(self):
        network_client = self.network_client
        results = []
        for lb in network_client.load_balancers.list(self.resource_group):
            if self.debug:
                print(lb)
            public_address = lb.frontend_ip_configurations[0].public_ip_address
            if public_address is not None:
                ip = network_client.public_ip_addresses.get(self.resource_group,
                                                            os.path.basename(public_address.id)).ip_address
            else:
                ip = lb.frontend_ip_configurations[0].private_ip_address
            dual_ip = None
            if len(lb.frontend_ip_configurations) > 1:
                dual_public_address = lb.frontend_ip_configurations[1].public_ip_address
                if dual_public_address is not None:
                    dual_ip = network_client.public_ip_addresses.get(self.resource_group,
                                                                     os.path.basename(public_address.id)).ip_address
                else:
                    dual_ip = lb.frontend_ip_configurations[1].private_ip_address
            protocol, ports, target = 'N/A', 'N/A', dual_ip or 'N/A'
            if lb.inbound_nat_rules or lb.load_balancing_rules:
                rule = lb.inbound_nat_rules[0] if lb.inbound_nat_rules else lb.load_balancing_rules[0]
                protocol = rule.protocol
                ports = rule.frontend_port_range_start if lb.inbound_nat_rules else rule.frontend_port
            results.append([lb.name, ip, protocol, ports, target])
        return results

    def add_vm_to_loadbalancer(self, vm, backend_id, ports, backend_id_dual=None):
        name = vm
        try:
            vm = self.compute_client.virtual_machines.get(self.resource_group, name)
        except:
            error("VM {name} not found")
            return {'result': 'success'}
        device = os.path.basename(vm.network_profile.network_interfaces[0].id)
        nic_data = self.network_client.network_interfaces.get(self.resource_group, device)
        nic_data.ip_configurations[0].load_balancer_backend_address_pools = [{'id': backend_id}]
        if backend_id_dual is not None:
            nic_data.ip_configurations[1].load_balancer_backend_address_pools = [{'id': backend_id_dual}]
        result = self.network_client.network_interfaces.begin_create_or_update(self.resource_group, device, nic_data)
        result.wait()
        ports = [22] + ports
        self.create_security_group(f'{name}-nsg', overrides={'ports': ports})
        return {'result': 'success'}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        try:
            subnet = ip_network(cidr, strict=False)
        except:
            return {'result': 'failure', 'reason': f"Invalid Cidr {cidr}"}
        subnet_ipv6 = str(subnet.version) == "6"
        dual_cidr = overrides.get('dual_cidr')
        if dual_cidr is not None:
            try:
                dual_network = ip_network(dual_cidr, strict=False)
            except:
                return {'result': 'failure', 'reason': f"Invalid Dual Cidr {cidr}"}
            dual_ipv6 = str(dual_network.version) == "6"
            if subnet_ipv6 == dual_ipv6:
                return {'result': 'failure', 'reason': "cidr and dual_cidr must be of different types"}
        data = {'address_prefix': cidr, 'address_prefixes': [cidr], 'tags': {'plan': plan}}
        if dual_cidr is not None:
            data['address_prefixes'].append(dual_cidr)
        network = overrides.get('network', name)
        if network not in self.list_networks():
            msg = f'Network {network} not found'
            return {'result': 'failure', 'reason': msg}
        if not nat and not subnet_ipv6:
            nat_gateway = f'nat-{network}'
            nat_gateway = self.network_client.nat_gateways.get(self.resource_group, nat_gateway)
            data['nat_gateway'] = {'id': nat_gateway.id}
        self.network_client.subnets.begin_create_or_update(self.resource_group, network, name, data)
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        subnets = self.list_subnets()
        if name not in subnets:
            msg = f'Subnet {name} not found'
            return {'result': 'failure', 'reason': msg}
        else:
            network = os.path.basename(subnets[name]['network'])
        result = self.network_client.subnets.begin_delete(self.resource_group, network, name)
        result.wait()
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        return [os.path.basename(z.id) for z in self.dns_client.zones.list()]

    def delete_identity(self, identity):
        self.msi_client.user_assigned_identities.delete(self.resource_group, identity)



from azure.identity import ClientSecretCredential
from azure.mgmt.containerservice import ContainerServiceClient
from kvirt.common import success, info2, pprint, error, get_ssh_pub_key, warning, fix_typos
import os
from yaml import safe_dump, safe_load

resource_group = 'kcli'


def project_init(config):
    admin_username = config.options.get('admin_username', 'superadmin')
    location = config.options.get('location', 'westus')
    subscription_id = config.options.get('subscription_id')
    app_id = config.options.get('app_id')
    tenant_id = config.options.get('tenant_id')
    secret = config.options.get('secret')
    return admin_username, location, subscription_id, app_id, tenant_id, secret


def get_kubeconfig(config, cluster, zonal=True):
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    results = containerservice_client.managed_clusters.list_cluster_admin_credentials(resource_group, cluster)
    kubeconfig = results.kubeconfigs[0].value
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    with open(f"{clusterdir}/auth/kubeconfig", 'wb') as f:
        f.write(kubeconfig)


def scale(config, plandir, cluster, overrides):
    data = {'workers': 2,
            'network': 'default',
            'disk_size': 0,
            'flavor': 'Standard_E2ds_v4',
            'autoscaling': True,
            'version': None}
    data.update(overrides)
    cluster = overrides.get('cluster') or cluster or 'myeks'
    version = data['version']
    workers = data['workers']
    disk_size = data['disk_size']
    flavor = data['flavor']
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    cluster_data = containerservice_client.managed_clusters.get(resource_group, cluster)
    pprint(f"Updating cluster {cluster}")
    if cluster_data.agent_pool_profiles.count != workers:
        cluster_data.agent_pool_profiles.count = workers
        cluster_data.agent_pool_profiles.min_count = workers
    if version is not None and cluster_data.kubernetes_version != version:
        cluster_data.kubernetes_version = version
    if flavor is not None and cluster_data.agent_pool_profiles.vm_size != flavor:
        cluster_data.agent_pool_profiles.vm_size = flavor
    if flavor is not None and cluster_data.agent_pool_profiles.osDiskSizeGB != disk_size:
        cluster_data.agent_pool_profiles.osDiskSizeGB = disk_size
    response = containerservice_client.managed_clusters.begin_create_or_update(resource_group, cluster, cluster_data)
    if config.debug:
        print(response.result())
    return {'result': 'success'}


def create(config, plandir, cluster, overrides, dnsconfig=None):
    data = safe_load(open(f'{plandir}/kcli_default.yml'))
    data.update(overrides)
    fix_typos(data)
    network = data['network']
    version = data['version']
    workers = data['workers']
    autoscaling = data['autoscaling']
    disk_size = data['disk_size']
    flavor = data['flavor']
    network_type = data['network_type']
    fips = data['fips']
    pub_key = data.get('pub_key') or get_ssh_pub_key()
    keys = data.get('keys', [])
    if pub_key is None:
        if keys:
            warning("Using first key from your keys array")
            pub_key = keys[0]
        else:
            msg = "No usable public key found, which is required for the deployment. Create one using ssh-keygen"
            return {'result': 'failure', 'reason': msg}
    pub_key = os.path.expanduser(pub_key)
    if not pub_key.startswith('ssh-') and os.path.exists(pub_key):
        pub_key = open(pub_key).read().strip()
    else:
        msg = f"Publickey file {pub_key} not found"
        return {'result': 'failure', 'reason': msg}
    plan = cluster
    tags = {'plan': cluster, 'kube': cluster, 'kubetype': 'aks'}
    cluster_data = {'name': cluster, 'tags': tags}
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if os.path.exists(clusterdir):
        return {'result': 'failure', 'reason': f"Remove existing directory {clusterdir} or use --force"}
    else:
        os.makedirs(clusterdir)
        os.mkdir(f"{clusterdir}/auth")
        with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
            installparam = overrides.copy()
            installparam['plan'] = plan
            installparam['cluster'] = cluster
            installparam['kubetype'] = 'aks'
            installparam['client'] = config.client
            safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    tags = {'plan': cluster, 'kube': cluster, 'kubetype': 'aks'}
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    agent_pool = {'name': cluster, 'vm_size': flavor, 'count': workers, 'min_count': workers, 'max_count': 30,
                  'enable_auto_scaling': autoscaling, 'mode': 'System', 'osDiskSizeGB': disk_size,
                  'enableFIPS': fips}
    if network is not None:
        agent_pool['vnetSubnetID'] = network
    service_principal_profile = {'client_id': app_id, 'secret': secret}
    linux_profile = {'admin_username': admin_username, 'ssh': {'public_keys': [{'key_data': pub_key}]}}
    cluster_data = {'location': location, 'dns_prefix': cluster, 'tags': tags,
                    'service_principal_profile': service_principal_profile,
                    'agent_pool_profiles': [agent_pool],
                    'linux_profile': linux_profile, 'enable_rbac': True}
    if version is not None:
        cluster_data['kubernetes_version'] = version
    network_profile = {}
    if network_type is not None:
        if network_type not in ['azure', 'kubenet']:
            return {'result': 'failure', 'reason': "Invalid network_type. Choose beetwen azure and kubenet"}
        network_profile['network_plugin'] = network_profile
    if network_profile:
        cluster_data['network_profile'] = network_profile
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    response = containerservice_client.managed_clusters.begin_create_or_update(resource_group, cluster, cluster_data)
    if config.debug:
        print(response.result())
    pprint(f"Waiting for cluster {cluster} to be created")
    response.wait()
    get_kubeconfig(config, cluster)
    success(f"Kubernetes cluster {cluster} deployed!!!")
    info2(f"export KUBECONFIG=$HOME/.kcli/clusters/{cluster}/auth/kubeconfig")
    info2("export PATH=$PWD:$PATH")
    return {'result': 'success'}


def delete(config, cluster, zonal=True):
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    try:
        response = containerservice_client.managed_clusters.begin_delete(resource_group, cluster)
        if config.debug:
            print(response.result())
        pprint(f"Waiting for cluster {cluster} to be deleted")
        response.wait()
    except Exception as e:
        error(f"Hit Issue when getting {cluster}: {e}")
        return {'result': 'failure', 'reason': 'Hit issue'}
    return {'result': 'success'}


def list(config):
    results = {}
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    response = containerservice_client.managed_clusters.list()
    for cluster in response:
        plan = cluster.tags.get('plan')
        results[cluster.name] = {'type': 'aks', 'plan': plan, 'vms': []}
    return results


def info(config, cluster, debug=False):
    results = {}
    admin_username, location, subscription_id, app_id, tenant_id, secret = project_init(config)
    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=app_id, client_secret=secret)
    containerservice_client = ContainerServiceClient(credential, subscription_id)
    try:
        response = containerservice_client.managed_clusters.get(resource_group, cluster)
        if debug:
            print(response)
    except Exception as e:
        error(e)
        return {}
    results = {'nodes': [], 'version': response.kubernetes_version}
    return results


def info_service(config, zonal=True):
    return {}



#!/usr/bin/env python
# coding=utf-8

import argparse
import json
import os
import sys


def error(text):
    color = "31"
    print(f'\033[0;{color}m{text}\033[0;0m')


def warning(text):
    color = "33"
    print(f'\033[0;{color}m{text}\033[0;0m')


def info(text):
    color = "36"
    print(f'\033[0;{color}m{text}\033[0;0m')


def success(text):
    color = "32"
    print(f'\033[0;{color}m{text}\033[0;0m')


def merge2ignitions(newdata, data):
    children = {'storage': 'files', 'passwd': 'users', 'systemd': 'units'}
    for key in children:
        childrenkey2 = 'path' if key == 'storage' else 'name'
        if key in data and key in newdata:
            if children[key] in data[key] and children[key] in newdata[key]:
                for entry in data[key][children[key]]:
                    if entry[childrenkey2] not in [x[childrenkey2] for x in newdata[key][children[key]]]:
                        newdata[key][children[key]].append(entry)
                    elif children[key] == 'users':
                        newusers = []
                        users = [x['name'] for x in data[key][children[key]] + newdata[key][children[key]]]
                        users = list(dict.fromkeys(users))
                        for user in users:
                            newuser = {'name': user}
                            sshkey1, sshkey2 = [], []
                            password = None
                            for y in data[key][children[key]]:
                                if y['name'] == user:
                                    sshkey1 = y['sshAuthorizedKeys'] if 'sshAuthorizedKeys' in y else []
                                    password = y.get('passwordHash')
                            for x in newdata[key][children[key]]:
                                if x['name'] == user:
                                    sshkey2 = x['sshAuthorizedKeys'] if 'sshAuthorizedKeys' in x else []
                                    password = x.get('passwordHash')
                            sshkeys = sshkey1
                            if sshkey2:
                                sshkeys.extend(sshkey2)
                            if sshkeys:
                                sshkeys = list(dict.fromkeys([sshkey.strip() for sshkey in sshkeys]))
                                newuser['sshAuthorizedKeys'] = sshkeys
                            if password is not None:
                                newuser['passwordHash'] = password
                            newusers.append(newuser)
                        newdata[key][children[key]] = newusers
            elif children[key] in data[key] and children[key] not in newdata[key]:
                newdata[key][children[key]] = data[key][children[key]]
        elif key in data and key not in newdata:
            newdata[key] = data[key]
    if 'ignition' in data and 'config' in data['ignition'] and data['ignition']['config']:
        newdata['ignition']['config'] = data['ignition']['config']
    return newdata


def mergeignition(args):
    version = '3.1.0'
    separators = (',', ':') if args.compact else (',', ': ')
    indent = None if args.compact else 4
    paths = args.paths
    data = {}
    for path in paths:
        if not os.path.exists(path):
            error(f"Missing path {path}. Ignoring")
            sys.exit(1)
        else:
            with open(path, 'r') as extra:
                try:
                    newdata = json.load(extra)
                except:
                    error(f"Couldn't process {path}. Leaving")
                    sys.exit(1)
            data = merge2ignitions(newdata, data)
    if 'ignition' not in data:
        data['ignition'] = {'config': {}, 'version': version}
    try:
        result = json.dumps(data, indent=indent, separators=separators, sort_keys=True)
    except:
        result = json.dumps(data, indent=indent, separators=separators)
    print(result)


def cli():
    parser = argparse.ArgumentParser(description="Merge your ignition files")
    parser.add_argument('-c', '--compact', action='store_true', help="Generate a compact ignition")
    parser.add_argument("paths", help="The path of your ignition files", nargs='+', type=str)
    parser.set_defaults(func=mergeignition)
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-

from kvirt import common
from kvirt.common import pprint, error, warning, get_ssh_pub_key
from kvirt.defaults import METADATA_FIELDS
from keystoneauth1 import loading
from keystoneauth1 import session
from glanceclient import Client as glanceclient
from cinderclient import client as cinderclient
from novaclient import client as novaclient
from neutronclient.v2_0.client import Client as neutronclient
import swiftclient.client as swiftclient
import os
from shutil import which
from time import sleep
import webbrowser
from ipaddress import ip_address, ip_network


class Kopenstack(object):
    """

    """
    def __init__(self, host='127.0.0.1', version='3', port=None, user='root', password=None, debug=False, project=None,
                 domain='Default', auth_url=None, ca_file=None, external_network=None, region_name=None,
                 glance_disk=False, auth_type='password', auth_token=None, application_credential_id=None,
                 application_credential_secret=None):
        self.debug = debug
        self.host = host
        loader = loading.get_plugin_loader(auth_type)
        if auth_type == 'password':
            auth = loader.load_from_options(auth_url=auth_url, username=user, password=password, project_name=project,
                                            user_domain_name=domain, project_domain_name=domain)
        elif auth_type == 'v3applicationcredential':
            auth = loader.load_from_options(auth_url=auth_url, application_credential_id=application_credential_id,
                                            application_credential_secret=application_credential_secret)
        elif auth_type == 'token':
            auth = loader.load_from_options(auth_url=auth_url, token=auth_token, project_id=project)
        else:
            error(f"Unsupported auth_type {auth_type}")
            self.conn = None
            return
        if ca_file is not None:
            cacert = os.path.expanduser(ca_file)
            sess_data = {'auth': auth, 'verify': cacert}
        else:
            cacert = None
            sess_data = {'auth': auth, 'verify': False}
        sess = session.Session(**sess_data)
        self.nova = novaclient.Client(version, session=sess, region_name=region_name)
        self.glance = glanceclient(version, session=sess, region_name=region_name)
        self.cinder = cinderclient.Client('3', session=sess, region_name=region_name)
        self.neutron = neutronclient(session=sess, region_name=region_name)
        os_options = {'user_domain_name': domain, 'project_domain_name': domain, 'project_name': project}
        self.swift = swiftclient.Connection(session=sess, os_options=os_options, cacert=cacert,
                                            insecure=ca_file is not None)
        self.conn = self.nova
        self.project = project
        self.external_network = external_network
        self.region_name = region_name
        self.glance_disk = glance_disk
        self.ca_file = ca_file

    def close(self):
        print("not implemented")

    def exists(self, name):
        return

    def net_exists(self, name):
        neutron = self.neutron
        networks = {net['name']: net['id'] for net in neutron.list_networks()['networks']}
        return name in networks

    def disk_exists(self, pool, name):
        print("not implemented")

    def create(self, name, virttype=None, profile='', plan='kvirt', flavor=None,
               cpumodel='host-model', cpuflags=[], cpupinning=[], numcpus=2, memory=512,
               guestid='guestrhel764', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True,
               diskinterface='virtio', nets=['default'], iso=None, vnc=True,
               cloudinit=True, reserveip=False, reservedns=False,
               reservehost=False, start=True, keys=[], cmds=[], ips=None,
               netmasks=None, gateway=None, nested=True, dns=None, domain=None,
               tunnel=False, files=[], enableroot=True, alias=[], overrides={},
               tags=[], storemetadata=False, sharedfolders=[], cmdline=None, placement=[], autostart=False,
               cpuhotplug=False, memoryhotplug=False, numamode=None, numa=[], pcidevices=[], tpm=False, rng=False,
               metadata={}, securitygroups=[], vmuser=None, guestagent=True):
        default_diskinterface = diskinterface
        glance = self.glance
        nova = self.nova
        neutron = self.neutron
        try:
            nova.servers.find(name=name)
            return {'result': 'failure', 'reason': f"VM {name} already exists"}
        except:
            pass
        allflavors = [f for f in nova.flavors.list()]
        allflavornames = [flavor.name for flavor in allflavors]
        if flavor is None:
            flavors = [flavor for flavor in allflavors if flavor.ram >= memory and flavor.vcpus >= numcpus]
            if flavors:
                flavor = flavors[0]
                pprint(f"Using flavor {flavor.name}")
            else:
                return {'result': 'failure', 'reason': "Couldn't find a valid flavor matching your specs"}
        elif flavor not in allflavornames:
            return {'result': 'failure', 'reason': f"Flavor {flavor} not found"}
        else:
            flavor = nova.flavors.find(name=flavor)
        nics = []
        need_floating = True
        for net in nets:
            if isinstance(net, str):
                netname = net
            elif isinstance(net, dict) and 'name' in net:
                netname = net['name']
            try:
                net = nova.neutron.find_network(name=netname)
                if net.to_dict()['router:external']:
                    need_floating = False
            except Exception as e:
                error(e)
                return {'result': 'failure', 'reason': f"Network {netname} not found"}
            nics.append({'net-id': net.id})
        target = iso or image
        if iso is not None and not self.glance_disk and len(disks) in [0, 1]:
            if len(disks) == 0:
                return {'result': 'failure', 'reason': "Booting from iso requires to specify at least one extra disk"}
            else:
                warning("Adding additional disk for booting from iso")
                disks.append(10)
        if target is not None:
            glanceimages = [img for img in glance.images.list() if img.name == target]
            if glanceimages:
                glanceimage = glanceimages[0]
            else:
                msg = f"you don't have image {target}"
                return {'result': 'failure', 'reason': msg}
        else:
            warning("a bootable disk is needed")
            glanceimage = None
        os_index = len(disks) - 1 if disks and iso is not None else 0
        block_device_mapping_v2 = []
        for index, disk in enumerate(disks):
            if index == os_index and self.glance_disk:
                continue
            diskname = f"{name}-disk{index}"
            letter = chr(index + ord('a'))
            diskinterface = default_diskinterface
            if isinstance(disk, int):
                disksize = disk
            elif isinstance(disk, str) and disk.isdigit():
                disksize = int(disk)
            elif isinstance(disk, dict):
                disksize = disk.get('size', '10')
                diskinterface = disk.get('interface', default_diskinterface)
            imageref = None
            if index == os_index and glanceimage is not None:
                imageref = glanceimage.id
                glanceimage = None
            newvol = self.cinder.volumes.create(name=diskname, size=disksize, imageRef=imageref)
            if index in [0, os_index]:
                self.cinder.volumes.set_bootable(newvol.id, True)
            while True:
                newvolstatus = self.cinder.volumes.get(newvol.id).status
                if newvolstatus == 'available':
                    break
                elif newvolstatus == 'error':
                    msg = f"Hit error when waiting for Disk {diskname} to be available"
                    return {'result': 'failure', 'reason': msg}
                else:
                    pprint(f"Waiting 10s for Disk {diskname} to be available")
                    sleep(10)
            block_device_mapping = {'device_name': f'vd{letter}', 'device_type': 'disk', 'disk_bus': diskinterface,
                                    'source_type': 'blank', "destination_type": "volume",
                                    'volume_size': disksize, "delete_on_termination": True, 'volume_id': newvol.id}
            if index == 0:
                block_device_mapping['boot_index'] = 0
            if iso is not None and index == os_index:
                block_device_mapping['device_type'] = 'cdrom'
                del block_device_mapping['disk_bus']
                block_device_mapping['boot_index'] = 1
            block_device_mapping_v2.append(block_device_mapping)
        key_name = 'kvirt'
        keypairs = [k.name for k in nova.keypairs.list()]
        if key_name not in keypairs:
            publickeyfile = get_ssh_pub_key()
            if publickeyfile is None:
                warning("neither id_rsa, id_dsa nor id_ed25519 public keys found in your .ssh directory, you "
                        "might have trouble accessing the vm")
            else:
                publickeyfile = open(publickeyfile).read()
                nova.keypairs.create(key_name, publickeyfile)
        elif keypairs:
            key_name = keypairs[0]
            if key_name != 'kvirt':
                pprint(f'Using keypair {key_name}')
        else:
            error("Couldn't locate or create keypair for use. Leaving...")
            return {'result': 'failure', 'reason': "No usable keypair found"}
        userdata = None
        if cloudinit:
            if image is not None and common.needs_ignition(image):
                version = common.ignition_version(image)
                userdata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                           domain=domain, files=files, enableroot=enableroot, overrides=overrides,
                                           version=version, plan=plan, image=image, vmuser=vmuser)
            else:
                userdata = common.cloudinit(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                            domain=domain, files=files, enableroot=enableroot, overrides=overrides,
                                            storemetadata=storemetadata, vmuser=vmuser)[0]
        meta = {x: metadata[x] for x in metadata if x in METADATA_FIELDS}
        instance = nova.servers.create(name=name, image=glanceimage, flavor=flavor, key_name=key_name, nics=nics,
                                       meta=meta, userdata=userdata, block_device_mapping_v2=block_device_mapping_v2,
                                       security_groups=securitygroups)
        tenant_id = instance.tenant_id
        if need_floating:
            floating_ips = [f['id'] for f in neutron.list_floatingips()['floatingips']
                            if f['port_id'] is None]
            if not floating_ips:
                network_id = None
                if self.external_network is not None:
                    networks = [n for n in neutron.list_networks()['networks'] if n['router:external']
                                if n['name'] == self.external_network]
                else:
                    networks = [n for n in neutron.list_networks()['networks'] if n['router:external']]
                if networks:
                    network_id = networks[0]['id']
                if network_id is not None and tenant_id is not None:
                    args = dict(floating_network_id=network_id, tenant_id=tenant_id)
                    floating_ip = neutron.create_floatingip(body={'floatingip': args})
                    floatingip_id = floating_ip['floatingip']['id']
                    floatingip_ip = floating_ip['floatingip']['floating_ip_address']
                    pprint(f'Assigning new floating ip {floatingip_ip} for this vm')
            else:
                floatingip_id = floating_ips[0]
            fixed_ip = None
            timeout = 0
            while fixed_ip is None:
                pprint("Waiting 5 seconds for vm to get an ip")
                sleep(5)
                timeout += 5
                if timeout >= 240:
                    error("Time out waiting for vm to get an ip")
                    break
                vm = nova.servers.get(instance.id)
                if vm.status.lower() == 'error':
                    msg = "Vm reports error status"
                    return {'result': 'failure', 'reason': msg}
                for key in list(vm.addresses):
                    entry1 = vm.addresses[key]
                    for entry2 in entry1:
                        if entry2['OS-EXT-IPS:type'] == 'fixed':
                            fixed_ip = entry2['addr']
                            break
            if fixed_ip is not None:
                fixedports = [i['id'] for i in neutron.list_ports()['ports']
                              if i['fixed_ips'] and i['fixed_ips'][0]['ip_address'] == fixed_ip]
                port_id = fixedports[0]
                neutron.update_floatingip(floatingip_id, {'floatingip': {'port_id': port_id}})
                if 'api_ip' in overrides:
                    api_ip = overrides['api_ip']
                    neutron.update_port(port_id, {'port': {'allowed_address_pairs': [{'ip_address': api_ip}]}})
            if not securitygroups:
                default_securitygroups = [s for s in neutron.list_security_groups()['security_groups']
                                          if s['name'] == 'default' and s['tenant_id'] == tenant_id]
                if default_securitygroups:
                    securitygroup = default_securitygroups[0]
                    securitygroupid = securitygroup['id']
                    sshrule = {'security_group_rule': {'direction': 'ingress', 'security_group_id': securitygroupid,
                                                       'port_range_min': '22', 'port_range_max': '22',
                                                       'protocol': 'tcp', 'remote_group_id': None,
                                                       'remote_ip_prefix': '0.0.0.0/0'}}
                    icmprule = {'security_group_rule': {'direction': 'ingress', 'security_group_id': securitygroupid,
                                                        'protocol': 'icmp', 'remote_group_id': None,
                                                        'remote_ip_prefix': '0.0.0.0/0'}}
                    try:
                        neutron.create_security_group_rule(sshrule)
                        neutron.create_security_group_rule(icmprule)
                    except:
                        pass
        if not start:
            vm.stop()
        return {'result': 'success'}

    def start(self, name):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm.start()
        return {'result': 'success'}

    def stop(self, name, soft=False):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm.stop()
        return {'result': 'success'}

    def create_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def delete_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def list_snapshots(self, base):
        print("not implemented")
        return []

    def revert_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def restart(self, name):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm.reboot()
        return {'result': 'success'}

    def info_host(self):
        print("not implemented")
        return {}

    def status(self, name):
        print("not implemented")
        return

    def list(self):
        vms = []
        nova = self.nova
        vmslist = nova.servers.list()
        for vm in vmslist:
            try:
                vms.append(self.info(vm.name, vm=vm))
            except:
                continue
        return sorted(vms, key=lambda x: x['name'])

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        url = vm.get_vnc_console('novnc')['console']['url']
        if web:
            return url
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{url}" if os.path.exists("/i_am_a_container") else url
            pprint(msg)
        else:
            pprint(f"Opening url: {url}")
            webbrowser.open(url, new=2, autoraise=True)
        return

    def serialconsole(self, name, web=False):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        cmd = vm.get_console_output()
        if web:
            return cmd
        print(cmd)

    def dnsinfo(self, name):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            return None, None
        dnsclient, domain = None, None
        metadata = vm.metadata
        if metadata is not None:
            if 'dnsclient' in metadata:
                dnsclient = metadata['dnsclient']
            if 'domain' in metadata:
                domain = metadata['domain']
        return dnsclient, domain

    def info(self, name, vm=None, debug=False):
        nova = self.nova
        cinder = self.cinder
        if vm is None:
            try:
                vm = nova.servers.find(name=name)
            except:
                error(f"VM {name} not found")
                return {}
        yamlinfo = {'name': vm.name, 'status': vm.status, 'project': self.project, 'id': vm.id}
        if vm.status.lower() == 'error':
            try:
                yamlinfo['error'] = vm.fault['message']
            except:
                pass

        source = ''
        if 'id' in vm.image:
            source = vm.image['id']
            try:
                source = self.glance.images.get(vm.image['id']).name
            except:
                pass
        yamlinfo['iso' if source.endswith('.iso') else 'image'] = source
        yamlinfo['creationdate'] = vm.created
        yamlinfo['user'] = common.get_user(source)
        flavor = nova.flavors.get(vm.flavor['id'])
        yamlinfo['flavor'] = flavor.name
        yamlinfo['memory'] = flavor.ram
        yamlinfo['numcpus'] = flavor.vcpus
        yamlinfo['nets'] = []
        index = 0
        ips = []
        for key in list(vm.addresses):
            entry1 = vm.addresses[key]
            for entry2 in entry1:
                mac = entry2['OS-EXT-IPS-MAC:mac_addr']
                if entry2['OS-EXT-IPS:type'] == 'floating':
                    publicip = entry2['addr']
                    yamlinfo['ip'] = publicip
                    ips.append(publicip)
                else:
                    net = {'device': f'eth{index}', 'mac': mac, 'net': key, 'type': entry2['addr']}
                    if index == 0:
                        yamlinfo['privateip'] = entry2['addr']
                    yamlinfo['nets'].append(net)
                    index += 1
        if 'ip' not in yamlinfo and 'privateip' in yamlinfo:
            privateip = yamlinfo['privateip']
            yamlinfo['ip'] = privateip
            ips.append(privateip)
        if len(ips) > 1:
            yamlinfo['ips'] = ips
        disks = []
        for disk in vm._info['os-extended-volumes:volumes_attached']:
            diskid = disk['id']
            try:
                volume = cinder.volumes.get(diskid)
            except cinderclient.exceptions.NotFound:
                disks.append({'device': 'N/A', 'size': 'N/A', 'format': 'N/A', 'type': 'N/A', 'path': diskid})
                continue
            disksize = volume.size
            devname = volume.name
            _type = volume.volume_type
            _format = volume.availability_zone
            volinfo = volume.to_dict()
            if 'volume_image_metadata' in volinfo and 'image_name' in volinfo['volume_image_metadata']:
                source = volinfo['volume_image_metadata']['image_name']
                yamlinfo['iso' if source.endswith('.iso') else 'image'] = source
            disks.append({'device': devname, 'size': disksize, 'format': _format, 'type': _type, 'path': diskid})
        if disks:
            yamlinfo['disks'] = disks
        metadata = vm.metadata
        if metadata is not None:
            for entry in metadata:
                yamlinfo[entry] = metadata[entry]
        if debug:
            yamlinfo['debug'] = str(vars(vm))
        return yamlinfo

    def ip(self, name):
        print("not implemented")

    def volumes(self, iso=False):
        images = []
        isos = []
        glance = self.glance
        for img in glance.images.list():
            imagename = img.name
            if imagename.endswith('.iso'):
                isos.append(imagename)
            else:
                images.append(imagename)
        if iso:
            return sorted(isos)
        else:
            return sorted(images)

    def delete(self, name, snapshots=False):
        cinder = self.cinder
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        floating_ips = {}
        try:
            fips = self.neutron.list_floatingips()
            floating_ips.update({f['floating_ip_address']: f['id'] for f in fips['floatingips']})
        except:
            # OVH does not provide floating ip networks, so the next error will occur
            # if some tries to get floating ip list:
            # neutronclient.common.exceptions.NotFound: The resource could not be found
            pass
        vm_floating_ips = []
        for key in list(vm.addresses):
            entry1 = vm.addresses[key]
            for entry2 in entry1:
                if entry2['OS-EXT-IPS:type'] == 'floating':
                    vm_floating_ips.append(entry2['addr'])
        vm.delete()

        # Wait until it's gone
        while True:
            try:
                nova.servers.get(vm.id)
                sleep(1)
            except novaclient.exceptions.NotFound:
                break

        for floating in vm_floating_ips:
            floatingid = floating_ips[floating]
            try:
                self.neutron.delete_floatingip(floatingid)
            except Exception as e:
                error(f"Hit {str(e)} when trying to delete floating {floating}")
        index = 0
        for disk in vm._info['os-extended-volumes:volumes_attached']:
            try:
                volume = cinder.volumes.get(disk['id'])
            except cinderclient.exceptions.NotFound:
                continue
            for attachment in volume.attachments:
                if attachment['server_id'] == vm.id:
                    cinder.volumes.detach(volume, attachment['attachment_id'])
            cinder.volumes.delete(disk['id'])
            index += 1
        return {'result': 'success'}

    def clone(self, old, new, full=False, start=False):
        print("not implemented")

    def update_metadata(self, name, metatype, metavalue, append=False):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return
        metadata = vm.metadata
        if append and metatype in metadata:
            metadata[metatype] += f",{metavalue}"
        else:
            metadata[metatype] = metavalue
        nova.servers.set_meta(vm.id, metadata)
        return {'result': 'success'}

    def update_memory(self, name, memory):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        currentflavor = nova.flavors.get(vm.flavor['id'])
        if currentflavor.ram >= int(memory):
            warning("No need to resize")
            return {'result': 'success'}
        allflavors = [f for f in nova.flavors.list() if f != currentflavor]
        flavors = [flavor for flavor in allflavors if flavor.ram >= int(memory) and flavor.vcpus >= currentflavor.vcpus]
        if flavors:
            flavor = flavors[0]
            pprint(f"Using flavor {flavor.name}")
            vm.resize(flavor.id)
            resizetimeout = 40
            resizeruntime = 0
            vmstatus = ''
            while vmstatus != 'VERIFY_RESIZE':
                if resizeruntime >= resizetimeout:
                    error("Time out waiting for resize to finish")
                    return {'result': 'failure', 'reason': "Time out waiting for resize to finish"}
                vm = nova.servers.find(name=name)
                vmstatus = vm.status
                sleep(2)
                pprint(f"Waiting for vm {name} to be in verify_resize")
                resizeruntime += 2
            vm.confirm_resize()
            return {'result': 'success'}
        else:
            error("Couldn't find matching flavor for this amount of memory")
            return {'result': 'failure', 'reason': "Couldn't find matching flavor for this amount of memory"}

    def update_flavor(self, name, flavor):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        currentflavor = nova.flavors.get(vm.flavor['id'])
        if currentflavor == flavor:
            return {'result': 'success'}
        flavors = [f for f in nova.flavors.list() if f.name == flavor]
        if not flavors:
            error(f"Flavor {flavor} doesn't exist")
            return {'result': 'failure', 'reason': f"Flavor {flavor} doesn't exist"}
        else:
            flavorid = flavors[0].id
            vm.resize(flavorid)
            resizetimeout = 40
            resizeruntime = 0
            vmstatus = ''
            while vmstatus != 'VERIFY_RESIZE':
                if resizeruntime >= resizetimeout:
                    error("Time out waiting for resize to finish")
                    return {'result': 'failure', 'reason': "Time out waiting for resize to finish"}
                vm = nova.servers.find(name=name)
                vmstatus = vm.status
                sleep(2)
                pprint(f"Waiting for vm {name} to be in verify_resize")
                resizeruntime += 2
            vm.confirm_resize()
            return {'result': 'success'}

    def update_cpus(self, name, numcpus):
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        currentflavor = nova.flavors.get(vm.flavor['id'])
        if currentflavor.vcpus >= numcpus:
            warning("No need to resize")
            return {'result': 'success'}
        allflavors = [f for f in nova.flavors.list() if f != currentflavor]
        flavors = [flavor for flavor in allflavors if flavor.ram >= currentflavor.ram and flavor.vcpus >= numcpus]
        if flavors:
            flavor = flavors[0]
            pprint(f"Using flavor {flavor.name}")
            vm.resize(flavor.id)
            resizetimeout = 40
            resizeruntime = 0
            vmstatus = ''
            while vmstatus != 'VERIFY_RESIZE':
                if resizeruntime >= resizetimeout:
                    error("Time out waiting for resize to finish")
                    return {'result': 'failure', 'reason': "Time out waiting for resize to finish"}
                vm = nova.servers.find(name=name)
                vmstatus = vm.status
                sleep(2)
                pprint(f"Waiting for vm {name} to be in verify_resize")
                resizeruntime += 2
            vm.confirm_resize()
            return {'result': 'success'}
        else:
            error("Couldn't find matching flavor for this number of cpus")
            return {'result': 'failure', 'reason': "Couldn't find matching flavor for this number of cpus"}

    def update_start(self, name, start=True):
        print("not implemented")

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)

    def update_iso(self, name, iso):
        if iso is not None:
            iso_images = [img for img in self.glance.images.list() if img.name == iso]
            if iso_images:
                iso_image = iso_images[0]
                iso_id = iso_image.id
            else:
                msg = f"Iso {iso} not found.Leaving..."
                error(msg)
                return {'result': 'failure', 'reason': msg}
        nova = self.nova
        cinder = self.cinder
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return
        currentiso = None
        if 'id' in vm.image:
            source = vm.image['id']
            try:
                source = self.glance.images.get(vm.image['id']).name
                if source.endswith('.iso'):
                    currentiso = source
                    warning("Update of iso set with glance are not supported")
                    return
            except:
                pass
        iso_name = ''
        for disk in vm._info['os-extended-volumes:volumes_attached']:
            diskid = disk['id']
            try:
                volume = cinder.volumes.get(diskid)
            except cinderclient.exceptions.NotFound:
                continue
            volinfo = volume.to_dict()
            if 'volume_image_metadata' in volinfo and 'image_name' in volinfo['volume_image_metadata']:
                source = volinfo['volume_image_metadata']['image_name']
                if source.endswith('.iso'):
                    currentiso = source
                    iso_name = volume.name
        if iso is None:
            if currentiso is None:
                return
            else:
                iso_volume = [volume for volume in cinder.volumes.list() if volume.name == iso_name][0]
                iso_volume.detach()
                iso_volume.delete()
        elif currentiso is not None and currentiso == iso:
            return
        else:
            volumes = [volume for volume in cinder.volumes.list() if volume.name == iso_name]
            if volumes:
                iso_volume = volumes[0]
                iso_volume.detach()
                iso_volume.delete()
            index = len(vm._info.get('os-extended-volumes:volumes_attached', []))
            diskname = f"{name}-disk{index}"
            letter = chr(index + ord('a'))
            volume = cinder.volumes.create(name=diskname, size=10, imageRef=iso_id)
            self.cinder.volumes.set_bootable(volume.id, True)
            cinder.volumes.attach(volume, vm.id, f'/dev/vd{letter}', mode='rw')
            return {'result': 'success'}

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        glance = self.glance
        cinder = self.cinder
        glanceimage = None
        if image is not None:
            glanceimages = [img for img in glance.images.list() if img.name == image]
            if glanceimages:
                glanceimage = glanceimages[0]
            else:
                msg = f"you don't have image {image}"
                return {'result': 'failure', 'reason': msg}
        cinder.volumes.create(name=name, size=size, imageRef=glanceimage)
        return {'result': 'success'}

    def add_disk(self, name, size, pool=None, thin=True, image=None,
                 shareable=False, existing=None, interface='virtio', novm=False, overrides={}, diskname=None):
        glance = self.glance
        cinder = self.cinder
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        glanceimage = None
        if image is not None:
            glanceimages = [img for img in glance.images.list() if img.name == image]
            if glanceimages:
                glanceimage = glanceimages[0]
            else:
                msg = f"you don't have image {image}"
                return {'result': 'failure', 'reason': msg}
        index = len(vm._info.get('os-extended-volumes:volumes_attached', []))
        diskname = f"{name}-disk{index}"
        letter = chr(index + ord('a'))
        volume = cinder.volumes.create(name=diskname, size=size, imageRef=glanceimage.id)
        cinder.volumes.attach(volume, vm.id, f'/dev/vd{letter}', mode='rw')
        return {'result': 'success'}

    def delete_disk(self, name=None, diskname=None, pool=None, novm=False):
        cinder = self.cinder
        nova = self.nova
        if name is None:
            volumes = [volume for volume in cinder.volumes.list() if volume.name == diskname]
            if volumes:
                volume = volumes[0]
            else:
                msg = f"Disk {diskname} not found"
                return {'result': 'failure', 'reason': msg}
            if volume.attachments:
                volume.detach()
            cinder.volumes.delete(volume.id)
            return {'result': 'success'}
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        for disk in vm._info['os-extended-volumes:volumes_attached']:
            volume = cinder.volumes.get(disk['id'])
            if diskname == volume.name:
                for attachment in volume.attachments:
                    if attachment['server_id'] == vm.id:
                        cinder.volumes.detach(volume, attachment['attachment_id'])
                cinder.volumes.delete(disk['id'])
                return {'result': 'success'}
        msg = f"Disk {diskname} not found in {name}"
        error(msg)
        return {'result': 'failure', 'reason': msg}

    def list_disks(self):
        volumes = {}
        cinder = self.cinder
        for volume in cinder.volumes.list():
            volumes[volume.name] = {'pool': volume.volume_type, 'path': volume.id}
        return volumes

    def add_nic(self, name, network, model='virtio'):
        print("not implemented")

    def delete_nic(self, name, interface):
        print("not implemented")

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")

    def delete_image(self, image, pool=None):
        pprint(f"Deleting image {image}")
        glance = self.glance
        for img in glance.images.list():
            if img.name == image:
                glance.images.delete(img.id)
                return {'result': 'success'}
        return {'result': 'failure', 'reason': f"Image {image} not found"}

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        downloaded = False
        shortimage = os.path.basename(url).split('?')[0]
        if name is not None and name.endswith('iso'):
            shortimage = name
        if [i for i in self.glance.images.list() if i['name'] == shortimage]:
            pprint(f"Image {shortimage} already there")
            return {'result': 'success'}
        if os.path.exists(url):
            pprint(f"Using {url} as path")
        elif not os.path.exists(f'/tmp/{shortimage}'):
            downloaded = True
            downloadcmd = f"curl -Lo /tmp/{shortimage} -f '{url}'"
            code = os.system(downloadcmd)
            if code != 0:
                return {'result': 'failure', 'reason': "Unable to download indicated image"}
        image_path = os.path.abspath(url) if os.path.exists(url) else f'/tmp/{shortimage}'
        need_uncompress = any(shortimage.endswith(suffix) for suffix in ['.gz', '.xz', '.bz2', '.zst'])
        if need_uncompress:
            extension = os.path.splitext(shortimage)[1].replace('.', '')
            executable = {'xz': 'unxz', 'gz': 'gunzip', 'bz2': 'bunzip2', 'zst': 'zstd'}
            flag = '--decompress' if extension == 'zstd' else '-f'
            executable = executable[extension]
            if which(executable) is not None:
                uncompresscmd = f"{executable} {flag} {image_path}"
                os.system(uncompresscmd)
            else:
                error(f"{executable} not found. Can't uncompress image")
                return {'result': 'failure', 'reason': f"{executable} not found. Can't uncompress image"}
            shortimage = shortimage.replace('.gz', '').replace('.xz', '').replace('.bz2', '').replace('.zst', '')
            image_path = image_path.replace('.gz', '').replace('.xz', '').replace('.bz2', '').replace('.zst', '')
        disk_format = 'iso' if shortimage.endswith('iso') else 'qcow2'
        if cmds:
            for cmd in cmds:
                pprint(f"Running {cmd} on {image_path}")
                if disk_format == 'iso':
                    if image_path not in cmd:
                        cmd += f" {image_path}"
                    os.system(cmd)
                elif which('virt-customize') is not None:
                    cmd = f"virt-customize -a {image_path} --run-command '{cmd}'"
                    os.system(cmd)
        glanceimage = self.glance.images.create(name=shortimage, disk_format=disk_format, container_format='bare')
        self.glance.images.upload(glanceimage.id, open(image_path, 'rb'))
        if downloaded:
            os.remove(f'/tmp/{shortimage}')
        return {'result': 'success'}

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        if nat:
            externalnets = [n for n in self.neutron.list_networks()['networks'] if n['router:external']]
            externalnet_id = externalnets[0]['id'] if externalnets else None
            routers = [router for router in self.neutron.list_routers()['routers'] if router['name'] == 'kvirt']
            router_id = routers[0]['id'] if routers else None
        try:
            ip_network(cidr)
        except:
            return {'result': 'failure', 'reason': f"Invalid Cidr {cidr}"}
        neutron = self.neutron
        network_id = None
        networks = {net['name']: net['id'] for net in neutron.list_networks()['networks']}
        if name not in networks:
            network = {'name': name, 'admin_state_up': True}
            if 'port_security_enabled' in overrides:
                network['port_security_enabled'] = bool(overrides['port_security_enabled'])
            network = neutron.create_network({'network': network})
            network_id = network['network']['id']
            tenant_id = network['network']['tenant_id']
        else:
            warning("Network already there. Creating subnet")
        if cidr is not None:
            if network_id is None:
                network_id = networks[name]
            cidrs = [s['cidr'] for s in neutron.list_subnets()['subnets'] if s['network_id'] == network_id]
            if cidr not in cidrs:
                subnet = {'name': cidr, 'network_id': network_id, 'ip_version': 4, "cidr": cidr, 'enable_dhcp': dhcp}
                if domain is not None:
                    subnet['dns_nameservers'] = [domain]
                subnet = neutron.create_subnet({'subnet': subnet})
                subnet_id = subnet['subnet']['id']
                tenant_id = subnet['subnet']['tenant_id']
            else:
                warning("Subnet already there. Leaving")
                return {'result': 'success'}
        if nat:
            if externalnet_id is not None:
                if router_id is None:
                    router = {'name': 'kvirt', 'tenant_id': tenant_id}
                    # router['external_gateway_info'] = {"network_id": externalnet_id, "enable_snat": True}
                    router = neutron.create_router({'router': router})
                    router_id = router['router']['id']
                    router_dict = {"network_id": externalnet_id}
                    neutron.add_gateway_router(router_id, router_dict)
                neutron.add_interface_router(router_id, {'subnet_id': subnet_id})
        return {'result': 'success'}

    def delete_network(self, name=None, cidr=None, force=False):
        neutron = self.neutron
        routers = [router for router in self.neutron.list_routers()['routers'] if router['name'] == 'kvirt']
        router_id = routers[0]['id'] if routers else None
        if router_id is not None:
            router = routers[0]
        networks = neutron.list_networks(name=name)
        if not networks:
            return {'result': 'failure', 'reason': f'Network {name} not found'}
        network_id = networks['networks'][0]['id']
        if cidr is None:
            ports = [p for p in neutron.list_ports()['ports']
                     if p['device_owner'] != 'network:router_interface' and network_id == network_id]
            if ports:
                return {'result': 'failure', 'reason': 'Non router ports still present in this network'}
            if router_id is not None:
                floating_ips = [f['id'] for f in neutron.list_floatingips()['floatingips']
                                if f['router_id'] == router_id]
                if floating_ips:
                    return {'result': 'failure', 'reason': 'Floating ips still in use through router on this network'}
                ports = [p for p in neutron.list_ports()['ports']
                         if p['device_id'] == router_id and network_id == network_id]
                routerports = len(ports)
                for port in ports:
                    neutron.remove_interface_router(router_id, {'port_id': port['id']})
                    routerports -= 1
            neutron.delete_network(network_id)
        else:
            subnets = [s['id'] for s in neutron.list_subnets()['subnets']
                       if s['network_id'] == network_id and s['cidr'] == cidr]
            if subnets:
                subnet_id = subnets[0]
                if router_id is not None:
                    floating_ips = [f['id'] for f in neutron.list_floatingips()['floatingips']
                                    if f['router_id'] == router_id]
                    if floating_ips:
                        return {'result': 'failure',
                                'reason': 'Floating ips still in use through router on this network'}
                    ports = [p for p in neutron.list_ports()['ports'] if p['device_id'] == router_id]
                    routerports = len(ports)
                    for port in ports:
                        if 'fixed_ips' in port and subnet_id in port['fixed_ips'][0].values():
                            neutron.remove_interface_router(router_id, {'port_id': port['id']})
                            routerports -= 1
                neutron.delete_subnet(subnet_id)
        if routerports == 0:
            if router['external_gateway_info']:
                neutron.remove_gateway_router(router_id)
            pprint("Removing unused router kvirt")
            neutron.delete_router(router_id)
        return {'result': 'success'}

    def list_pools(self):
        print("not implemented")
        return []

    def list_networks(self):
        networks = {}
        neutron = self.neutron
        for subnet in neutron.list_subnets()['subnets']:
            subnetname = subnet['name']
            subnet_id = subnet['id']
            cidr = subnet['cidr']
            dhcp = subnet['enable_dhcp']
            network_id = subnet['network_id']
            network = neutron.show_network(network_id)
            mode = 'external' if network['network']['router:external'] else 'isolated'
            # networks = [n for n in neutron.list_networks()['networks'] if n['router:external']]
            networkname = neutron.show_network(network_id)['network']['name']
            ports = [p for p in neutron.list_ports()['ports']
                     if p['device_owner'] == 'network:router_interface' and network_id == network_id]
            for port in ports:
                if 'fixed_ips' in port and subnet_id in port['fixed_ips'][0].values():
                    mode = 'nat'
                    break
            if networkname in networks:
                networks[networkname]['domain'] = f"{networks[networkname]['domain']}, {subnetname}"
            else:
                networks[networkname] = {'cidr': cidr, 'dhcp': dhcp, 'domain': subnetname, 'type': 'routed',
                                         'mode': mode}
        return networks

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        if self.debug and networkinfo:
            network = [net for net in self.neutron.list_networks()['networks'] if net['name'] == name][0]
            print(network)
            subnets = [sub for sub in self.neutron.list_subnets()['subnets'] if sub['id'] in network['subnets']]
            for subnet in subnets:
                print(subnet)
        return networkinfo

    def list_subnets(self):
        print("not implemented")
        return {}

    def delete_pool(self, name, full=False):
        print("not implemented")

    def network_ports(self, name):
        print("not implemented")

    def vm_ports(self, name):
        return ['default']

    def get_pool_path(self, pool):
        print("not implemented")

    def list_flavors(self):
        nova = self.nova
        nova.flavors.list
        flavors = [[flavor.name, flavor.vcpus, flavor.ram] for flavor in nova.flavors.list()]
        return flavors

    def export(self, name, image=None):
        cinder = self.cinder
        nova = self.nova
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        for disk in vm._info['os-extended-volumes:volumes_attached']:
            volume = cinder.volumes.get(disk['id'])
            for attachment in volume.attachments:
                newname = image if image is not None else volume.name.replace('-disk0', '')
                volume.upload_to_image(True, newname, 'bare', 'qcow2')
                status = ''
                timeout = 0
                while status != 'available':
                    status = cinder.volumes.get(disk['id']).status
                    pprint("Waiting 5 seconds for export to complete")
                    sleep(5)
                    timeout += 5
                    if timeout >= 90:
                        error("Time out waiting for export to complete")
                        break
                break
        return {'result': 'success'}

    def list_dns(self, domain):
        return []

    def create_network_port(self, name, network, ip=None, floating=False, security=True):
        neutron = self.neutron
        matchingports = [i for i in neutron.list_ports()['ports'] if i['name'] == name]
        if matchingports:
            msg = f"Port {name} already exists"
            pprint(msg)
            return {'result': 'success'}
        networks = [net for net in neutron.list_networks()['networks'] if net['name'] == network]
        if not networks:
            msg = f"Network {network} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        else:
            network = networks[0]
        network_id = network['id']
        port = {'name': name, "admin_state_up": True, "network_id": network_id, 'port_security_enabled': security}
        if ip is not None:
            for subnet in neutron.list_subnets()['subnets']:
                subnet_name = subnet['name']
                subnet_id = subnet['id']
                cidr = subnet['cidr']
                if network_id == subnet['network_id'] and ip_address(ip) in ip_network(cidr):
                    msg = f"Using matching subnet {subnet_name} with cidr {cidr}"
                    pprint(msg)
                    port['fixed_ips'] = [{'ip_address': ip, 'subnet_id': subnet_id}]
        result = neutron.create_port({'port': port})
        port_id = result['port']['id']
        result = {'result': 'success'}
        if floating:
            tenant_id = network['tenant_id']
            if self.external_network is not None:
                network_id = self.external_network
                external_networks = [n for n in neutron.list_networks()['networks'] if n['router:external']
                                     if n['name'] == self.external_network]
            else:
                external_networks = [n for n in neutron.list_networks()['networks'] if n['router:external']]
            if external_networks:
                network_id = external_networks[0]['id']
            else:
                msg = "No valid external network found for floating ips"
                error(msg)
                return {'result': 'failure', 'reason': msg}
            args = dict(floating_network_id=network_id, tenant_id=tenant_id, port_id=port_id)
            floating_ip = neutron.create_floatingip(body={'floatingip': args})
            floatingip_ip = floating_ip['floatingip']['floating_ip_address']
            pprint(f'Assigning new floating ip {floatingip_ip} for this port')
            result['floating'] = floatingip_ip
        return result

    def delete_network_port(self, name):
        neutron = self.neutron
        matchingports = [i for i in neutron.list_ports()['ports'] if i['name'] == name]
        if not matchingports:
            msg = f"Port {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        self.neutron.delete_port(matchingports[0]['id'])

    def create_bucket(self, bucket, public=False):
        swift = self.swift
        if bucket in self.list_buckets():
            error(f"Bucket {bucket} already exists")
            return
        headers = {"X-Container-Read": ".r:*"} if public else {}
        swift.put_container(bucket, headers=headers)

    def delete_bucket(self, bucket):
        swift = self.swift
        try:
            containerinfo = swift.get_container(bucket)
        except:
            error(f"Inexistent bucket {bucket}")
            return
        for obj in containerinfo[1]:
            obj_name = obj['name']
            pprint(f"Deleting object {obj_name} in bucket {bucket}")
            swift.delete_object(bucket, obj_name)
        swift.delete_container(bucket)

    def delete_from_bucket(self, bucket, path):
        swift = self.swift
        try:
            containerinfo = swift.get_container(bucket)
        except:
            error(f"Inexistent bucket {bucket}")
            return
        for obj in containerinfo[1]:
            obj_name = obj['name']
            if path == obj_name:
                swift.delete_object(bucket, obj_name)
                break

    def download_from_bucket(self, bucket, path):
        swift = self.swift
        try:
            resp_headers, obj_contents = swift.get_object(bucket, path)
        except Exception as e:
            error(f"Got {e}")
            return
        with open(path, 'wb') as f:
            f.write(obj_contents)

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        swift = self.swift
        if not os.path.exists(path):
            error(f"Invalid path {path}")
            return
        dest = os.path.basename(path)
        with open(path, 'rb') as f:
            swift.put_object(bucket, dest, contents=f, content_type='text/plain')
        if public:
            headers = {"X-Container-Read": ".r:*"}
            swift.post_container(bucket, headers=headers)

    def list_buckets(self):
        swift = self.swift
        return [container['name'] for container in swift.get_account()[1]]

    def list_bucketfiles(self, bucket):
        swift = self.swift
        try:
            containerinfo = swift.get_container(bucket)
        except:
            error(f"Inexistent bucket {bucket}")
            return []
        return [obj['name'] for obj in containerinfo[1]]

    def public_bucketfile_url(self, bucket, path):
        swift_url = self.swift.http_connection()[0].geturl()
        return f"{swift_url}/{bucket}/{path}"

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False):
        print("not implemented")

    def update_nic(self, name, index, network):
        print("not implemented")

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        neutron = self.neutron
        networks = [net for net in neutron.list_networks()['networks'] if net['name'] == name]
        if not networks:
            msg = f"Network {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        else:
            network = networks[0]
            network_id = network['id']
        network_data = {}
        if 'port_security_enabled' in overrides and isinstance(overrides['port_security_enabled'], bool):
            network_data['port_security_enabled'] = overrides['port_security_enabled']
        if 'mtu' in overrides and isinstance(overrides['port_security_enabled'], int):
            network_data['mtu'] = overrides['mtu']
        if 'description' in overrides:
            network_data['description'] = str(overrides['description'])
        if 'tags' in overrides and isinstance(overrides['tags'], list):
            network_data['tags'] = overrides['tags']
        if 'availability_zones' in overrides and isinstance(overrides['availability_zones'], list):
            network_data['availability_zones'] = overrides['availability_zones']
        if 'availability_zone_hints' in overrides and isinstance(overrides['availability_zone_hints'], list):
            network_data['availability_zone_hints'] = overrides['availability_zone_hints']
        if network_data:
            neutron.update_network(network_id, {'network': network_data})
        subnets = [sub for sub in self.neutron.list_subnets()['subnets'] if sub['id'] in network['subnets']]
        if subnets:
            subnet_data = {}
            subnet = subnets[0]
            subnet_id = subnet['id']
            currentdhcp = subnet['enable_dhcp']
            if dhcp is not None:
                if not dhcp and currentdhcp:
                    subnet_data['enable_dhcp'] = False
                if dhcp and not currentdhcp:
                    subnet_data['enable_dhcp'] = True
            if 'dns_nameservers' in overrides and isinstance(overrides['dns_nameservers'], list):
                subnet_data['dns_nameservers'] = overrides['dns_nameservers']
            if 'dns' in overrides and isinstance(overrides['dns'], list):
                subnet_data['dns_nameservers'] = overrides['dns']
            if subnet_data:
                neutron.update_subnet(subnet_id, {'subnet': subnet_data})
        return {'result': 'success'}

    def list_security_groups(self, network=None):
        neutron = self.neutron
        return [s['name'] for s in neutron.list_security_groups()['security_groups']]

    def create_security_group(self, name, overrides={}):
        neutron = self.neutron
        security_group = {'name': name}
        sg = neutron.create_security_group({'security_group': security_group})
        sgid = sg['security_group']['id']
        icmprule = {'security_group_rule': {'direction': 'ingress', 'security_group_id': sgid,
                                            'protocol': 'icmp', 'remote_group_id': None,
                                            'remote_ip_prefix': '0.0.0.0/0'}}
        neutron.create_security_group_rule(icmprule)
        ports = overrides.get('ports', [])
        for port in ports:
            if isinstance(port, str) or isinstance(port, int):
                protocol = 'tcp'
                fromport, toport = port, port
            elif isinstance(port, dict):
                protocol = port.get('protocol', 'tcp')
                fromport = port.get('from')
                toport = port.get('to') or fromport
                if fromport is None:
                    warning(f"Missing from in {ports}. Skipping")
                    continue
            pprint(f"Adding rule from {fromport} to {toport} protocol {protocol}")
            rule = {'security_group_rule': {'direction': 'ingress', 'security_group_id': sgid,
                                            'port_range_min': str(fromport), 'port_range_max': str(toport),
                                            'protocol': protocol, 'remote_group_id': None,
                                            'remote_ip_prefix': '0.0.0.0/0'}}
            neutron.create_security_group_rule(rule)
        return {'result': 'success'}

    def delete_security_group(self, name):
        sgs = [sg for sg in self.neutron.list_security_groups()['security_groups'] if sg['name'] == name]
        if sgs:
            sgs[0].delete()
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def provider_network(self, name):
        net = self.nova.neutron.find_network(name=name)
        return True if net.to_dict()['router:external'] else False

    def info_subnet(self, name):
        print("not implemented")
        return {}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        print("not implemented")
        return []

    def detach_disks(self, name):
        nova = self.nova
        cinder = self.cinder
        try:
            vm = nova.servers.find(name=name)
        except:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        all_disks = vm._info['os-extended-volumes:volumes_attached']
        vm_disks = all_disks[1:] if len(all_disks) > 1 else []
        for disk in vm_disks:
            volume = cinder.volumes.get(disk['id'])
            for attachment in volume.attachments:
                if attachment['server_id'] == vm.id:
                    cinder.volumes.detach(volume, attachment['attachment_id'])
        return {'result': 'success'}



#!/usr/bin/env python
# coding=utf-8

# import base64
from kvirt.config import Kconfig
import os
import json
import sys
import google.auth
import google.auth.transport.requests


def cli():
    if len(sys.argv) != 2:
        print("Usage gketoken $client")
        sys.exit(1)
    client = sys.argv[1]
    config = Kconfig(client)
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = os.path.expanduser(config.options.get('credentials'))
    credentials, your_project_id = google.auth.default(scopes=["https://www.googleapis.com/auth/cloud-platform"])
    auth_req = google.auth.transport.requests.Request()
    credentials.refresh(auth_req)
    token = credentials.token
    result = {'apiVersion': 'client.authentication.k8s.io/v1beta1', 'kind': 'ExecCredential', 'spec': {},
              'status': {'expirationTimestamp': '2030-10-01T15:05:17Z', 'token': token}}
    print(json.dumps(result))


if __name__ == '__main__':
    cli()



from ipaddress import ip_network
from kvirt.common import success, pprint, warning, info2, container_mode, wait_cloud_dns, update_etc_hosts, fix_typos
from kvirt.common import get_kubectl, get_ssh_pub_key, _ssh_credentials, ssh, deploy_cloud_storage, wait_for_nodes
from kvirt.defaults import UBUNTUS
import os
from random import choice
from re import match
from shutil import which
from string import ascii_lowercase, ascii_letters, digits
from subprocess import call
from tempfile import NamedTemporaryFile
from yaml import safe_dump, safe_load

cloud_providers = ['aws', 'azure', 'gcp', 'ibm', 'hcloud']


def scale(config, plandir, cluster, overrides):
    storedparameters = overrides.get('storedparameters', True)
    plan = cluster
    data = {'cluster': cluster, 'sslip': False, 'kube': cluster, 'kubetype': 'rke2', 'image': 'centos9stream',
            'extra_scripts': []}
    data['basedir'] = '/workdir' if container_mode() else '.'
    cluster = data.get('cluster')
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if not os.path.exists(clusterdir):
        warning(f"Creating {clusterdir} from your input (auth creds will be missing)")
        data['client'] = config.client
        overrides['cluster'] = cluster
        first_vm = f"{cluster}-ctlplane-0"
        first_ip, first_vmport = _ssh_credentials(config.k, first_vm)[1:]
        data['first_ip'] = first_ip
        domain = overrides.get('domain')
        if domain is None:
            domaincmd = "grep DOMAIN= /root/bootstrap.sh"
            domaincmd = ssh(first_vm, ip=first_ip, user='root', tunnel=config.tunnel,
                            tunnelhost=config.tunnelhost, tunnelport=config.tunnelport, tunneluser=config.tunneluser,
                            insecure=True, cmd=domaincmd, vmport=first_vmport)
            data['domain'] = os.popen(domaincmd).read().strip().split('=')[1]
        os.mkdir(clusterdir)
        tokencmd = "grep TOKEN= /root/bootstrap.sh"
        tokencmd = ssh(first_vm, ip=first_ip, user='root', tunnel=config.tunnel,
                       tunnelhost=config.tunnelhost, tunnelport=config.tunnelport, tunneluser=config.tunneluser,
                       insecure=True, cmd=tokencmd, vmport=first_vmport)
        data['token'] = os.popen(tokencmd).read().strip().split('=')[1]
    if 'api_ip' not in data:
        data['api_ip'] = None
    if 'first_ip' not in data:
        first_info = config.k.info(f'{cluster}-ctlplane-0')
        first_ip = first_info.get('private_ip') or first_info.get('ip')
        data['first_ip'] = first_ip
    if storedparameters and os.path.exists(f"{clusterdir}/kcli_parameters.yml"):
        with open(f"{clusterdir}/kcli_parameters.yml", 'r') as install:
            installparam = safe_load(install)
            data.update(installparam)
            plan = installparam.get('plan', plan)
    data.update(overrides)
    data['scale'] = True
    if os.path.exists(clusterdir):
        with open(f"{clusterdir}/kcli_parameters.yml", 'w') as paramfile:
            safe_dump(data, paramfile)
    client = config.client
    pprint(f"Scaling on client {client}")
    image = data.get('image')
    if 'ubuntu' not in data:
        data['ubuntu'] = 'ubuntu' in image.lower() or len([u for u in UBUNTUS if u in image]) > 0
    os.chdir(os.path.expanduser("~/.kcli"))
    for role in ['ctlplanes', 'workers']:
        overrides = data.copy()
        if overrides.get(role, 0) == 0:
            continue
        threaded = data.get('threaded', False) or data.get(f'{role}_threaded', False)
        result = config.plan(plan, inputfile=f'{plandir}/{role}.yml', overrides=overrides, threaded=threaded)
        if result['result'] != 'success':
            return result
        elif result.get('newvms', []):
            pprint(f"{role.capitalize()} nodes will join the cluster in a few minutes")
    return {'result': 'success'}


def create(config, plandir, cluster, overrides):
    provider = config.type
    k = config.k
    data = safe_load(open(f'{plandir}/kcli_default.yml'))
    async_install = data['async']
    data.update(overrides)
    fix_typos(data)
    if 'keys' not in overrides and get_ssh_pub_key() is None:
        msg = "No usable public key found, which is required for the deployment. Create one using ssh-keygen"
        return {'result': 'failure', 'reason': msg}
    valid_sdns = ['none', 'canal', 'calico', 'cilium']
    if data['sdn'].lower() not in valid_sdns:
        msg = f"Invalid sdn, Choose between {','.join(valid_sdns)}"
        return {'result': 'failure', 'reason': msg}
    data['cluster'] = cluster
    plan = cluster
    data['kube'] = data['cluster']
    data['kubetype'] = 'rke2'
    autolabel = data['autolabel']
    cloud_lb = data['cloud_lb']
    cloud_dns = data['cloud_dns']
    cloud_storage = data['cloud_storage']
    autoscale = data.get('autoscale')
    ctlplanes = data.get('ctlplanes', 1)
    if ctlplanes == 0:
        msg = "Invalid number of ctlplanes"
        return {'result': 'failure', 'reason': msg}
    if ctlplanes > 1 and provider in cloud_providers and not cloud_lb:
        msg = "Multiple ctlplanes require cloud_lb to be set to True"
        return {'result': 'failure', 'reason': msg}
    network = data.get('network', 'default')
    api_ip = data.get('api_ip')
    if provider in cloud_providers:
        domain = data.get('domain', 'karmalabs.corp')
        api_ip = f"{cluster}-ctlplane.{domain}"
    elif api_ip is None:
        networkinfo = k.info_network(network)
        if not networkinfo:
            msg = f"Issue getting network {network}"
            return {'result': 'failure', 'reason': msg}
        if provider == 'kvm' and networkinfo['type'] == 'routed':
            cidr = networkinfo['cidr']
            if cidr == 'N/A':
                msg = "Couldnt gather an api_ip from your specified network"
                return {'result': 'failure', 'reason': msg}
            api_index = 2 if ':' in cidr else -3
            api_ip = str(ip_network(cidr)[api_index])
            warning(f"Using {api_ip} as api_ip")
            data['api_ip'] = api_ip
        elif provider == 'kubevirt':
            selector = {'kcli/plan': plan, 'kcli/role': 'ctlplane'}
            service_type = "LoadBalancer" if k.access_mode == 'LoadBalancer' else 'ClusterIP'
            api_ip = config.k.create_service(f"{cluster}-api", config.k.namespace, selector, _type=service_type,
                                             ports=[6443])
            if api_ip is None:
                msg = "Couldnt get an kubevirt api_ip from service"
                return {'result': 'failure', 'reason': msg}
            else:
                pprint(f"Using api_ip {api_ip}")
                data['api_ip'] = api_ip
        else:
            msg = "You need to define api_ip in your parameters file"
            return {'result': 'failure', 'reason': msg}
    if provider not in cloud_providers:
        if data.get('virtual_router_id') is None:
            data['virtual_router_id'] = hash(data['cluster']) % 254 + 1
        virtual_router_id = data['virtual_router_id']
        pprint(f"Using keepalived virtual_router_id {virtual_router_id}")
        auth_pass = ''.join(choice(ascii_letters + digits) for i in range(5))
        data['auth_pass'] = auth_pass
    valid_characters = ascii_lowercase + digits
    token = data['token']
    if token is not None:
        if match(r"[a-z0-9]{6}.[a-z0-9]{16}", data['token']) is not None:
            msg = "Incorrect token format. It should be [a-z0-9]{6}.[a-z0-9]{16}"
            return {'result': 'failure', 'reason': msg}
    else:
        token1 = ''.join(choice(valid_characters) for i in range(6))
        token2 = ''.join(choice(valid_characters) for i in range(16))
        token = f'{token1}.{token2}'
        data['token'] = token
    data['basedir'] = '/workdir' if container_mode() else '.'
    cluster = data.get('cluster')
    image = data.get('image', 'centos9stream')
    data['ubuntu'] = 'ubuntu' in image.lower() or len([u for u in UBUNTUS if u in image]) > 0
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if os.path.exists(clusterdir):
        msg = f"Remove existing directory {clusterdir} or use --force"
        return {'result': 'failure', 'reason': msg}
    if which('kubectl') is None:
        get_kubectl()
    if not os.path.exists(clusterdir):
        os.makedirs(clusterdir)
        os.mkdir(f"{clusterdir}/auth")
    result = config.plan(plan, inputfile=f'{plandir}/bootstrap.yml', overrides=data)
    if result['result'] != "success":
        return result
    first_info = config.k.info(f'{cluster}-ctlplane-0')
    first_ip = first_info.get('private_ip') or first_info.get('ip')
    data['first_ip'] = first_ip
    with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
        installparam = overrides.copy()
        installparam['api_ip'] = api_ip
        if 'virtual_router_id' in data:
            installparam['virtual_router_id'] = data['virtual_router_id']
        if 'auth_pass' in data:
            installparam['auth_pass'] = auth_pass
        installparam['client'] = config.client
        installparam['plan'] = plan
        installparam['token'] = token
        installparam['cluster'] = cluster
        installparam['kubetype'] = 'rke2'
        installparam['image'] = image
        installparam['ubuntu'] = 'ubuntu' in image.lower() or len([u for u in UBUNTUS if u in image]) > 1
        installparam['first_ip'] = first_ip
        safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    if ctlplanes > 1:
        ctlplane_threaded = data.get('threaded', False) or data.get('ctlplanes_threaded', False)
        result = config.plan(plan, inputfile=f'{plandir}/ctlplanes.yml', overrides=data, threaded=ctlplane_threaded)
        if result['result'] != "success":
            return result
    if cloud_lb and provider in cloud_providers:
        config.k.delete_dns(f'api.{cluster}', domain=domain)
        if provider == 'aws':
            data['vpcid'] = config.k.get_vpcid_of_vm(f"{cluster}-ctlplane-0")
        result = config.plan(plan, inputfile=f'{plandir}/cloud_lb_api.yml', overrides=data)
        if result['result'] != 'success':
            return result
    workers = data.get('workers', 0)
    if workers > 0:
        pprint("Deploying workers")
        if 'name' in data:
            del data['name']
        os.chdir(os.path.expanduser("~/.kcli"))
        worker_threaded = data.get('threaded', False) or data.get('workers_threaded', False)
        config.plan(plan, inputfile=f'{plandir}/workers.yml', overrides=data, threaded=worker_threaded)
    if async_install:
        success(f"Kubernetes cluster {cluster} deployed!!!")
        info2(f"get kubeconfig from {cluster}-ctlplane-0 /root")
        return {'result': 'success'}
    success(f"Kubernetes cluster {cluster} deployed!!!")
    info2(f"export KUBECONFIG=$HOME/.kcli/clusters/{cluster}/auth/kubeconfig")
    info2("export PATH=$PWD:$PATH")
    if provider in cloud_providers and cloud_lb:
        if cloud_dns:
            wait_cloud_dns(cluster, domain)
        # elif api_ip is not None and api_ip == f'api.{cluster}.{domain}':
        else:
            for lbentry in config.list_loadbalancers():
                if lbentry[0] == f'api-{cluster}':
                    lb_ip = lbentry[1]
                    update_etc_hosts(cluster, domain, lb_ip)
                    break
    os.environ['KUBECONFIG'] = f"{clusterdir}/auth/kubeconfig"
    if ctlplanes + workers > 1:
        ready = wait_for_nodes(ctlplanes + workers)
        if not ready:
            msg = "Timeout waiting for all nodes to join"
            return {'result': 'failure', 'reason': msg}
    if autoscale:
        config.import_in_kube(network=network, secure=True)
        with NamedTemporaryFile(mode='w+t') as temp:
            commondir = os.path.dirname(pprint.__code__.co_filename)
            autoscale_overrides = {'cluster': cluster, 'kubetype': 'rke2', 'workers': workers, 'replicas': 1}
            autoscale_data = config.process_inputfile(cluster, f"{commondir}/autoscale.yaml.j2",
                                                      overrides=autoscale_overrides)
            temp.write(autoscale_data)
            autoscalecmd = f"kubectl create -f {temp.name}"
            call(autoscalecmd, shell=True)
    if provider in cloud_providers and cloud_storage:
        if provider == 'aws':
            pprint("Deploying cloud storage class")
            deploy_cloud_storage(config, cluster)
    if autolabel:
        autolabelcmd = 'kubectl apply -f https://raw.githubusercontent.com/karmab/autolabeller/main/autorules.yml'
        call(autolabelcmd, shell=True)
    return {'result': 'success'}
